#!/usr/bin/env python3
# helper script to run and save benchmarks against named refs.
# AKA a shitty version of asv's env management
# FIXME:
# 1. The Icechunk Spec Version is taken from the running env. This is wrong :(

import argparse
import glob
import os
import subprocess
import tempfile
import tomllib
from collections.abc import Callable
from functools import partial

import tqdm
import tqdm.contrib.concurrent
from helpers import (
    assert_cwd_is_icechunk_python,
    get_coiled_kwargs,
    get_full_commit,
    rdms,
    setup_logger,
)

logger = setup_logger()

PIP_OPTIONS = "--disable-pip-version-check -q"
PYTEST_OPTIONS = "-q --durations 10 --rootdir=benchmarks --tb=line"
TMP = tempfile.gettempdir()
CURRENTDIR = os.getcwd()


assert_cwd_is_icechunk_python()


def process_map(*, serial: bool, func: Callable, iterable):
    if serial:
        for i in iterable:
            func(i)
    else:
        tqdm.contrib.concurrent.process_map(func, iterable)


def get_benchmark_deps(filepath: str) -> str:
    """needed since
    1. benchmark deps may have changed in the meantime.
    2. we can't specify optional extras when installing from a subdirectory
       https://pip.pypa.io/en/stable/topics/vcs-support/#url-fragments
    """
    with open(filepath, mode="rb") as f:
        data = tomllib.load(f)
    return (
        " ".join(data["project"]["optional-dependencies"].get("benchmark", ""))
        + " "
        + " ".join(data["project"]["optional-dependencies"].get("test", ""))
    )


class Runner:
    bench_store_dir = None

    def __init__(self, *, ref: str, where: str, save_prefix: str) -> None:
        self.ref = ref
        self.full_commit = get_full_commit(ref)
        self.commit = self.full_commit[:8]
        self.where = where
        self.save_prefix = save_prefix
        # shorten the name so `pytest-benchmark compare` is readable
        self.clean_ref = self.ref.removeprefix("icechunk-v0.1.0-alph")

    @property
    def pip_github_url(self) -> str:
        # optional extras cannot be specified here, "not guaranteed to work"
        # https://pip.pypa.io/en/stable/topics/vcs-support/#url-fragments
        return f"git+https://github.com/earth-mover/icechunk.git@{self.full_commit}#subdirectory=icechunk-python"

    @property
    def prefix(self) -> str:
        # try:
        #     return f"v{ic.spec_version():02d}"
        # except AttributeError:
        return f"{self.ref}_{self.commit}"

    @property
    def ref_commit(self) -> str:
        return f"{self.ref}_{self.commit}"

    def sync_benchmarks_folder(self) -> None:
        """Sync the benchmarks folder over to the cwd."""
        raise NotImplementedError

    def execute(cmd: str) -> None:
        """Execute a command"""
        raise NotImplementedError

    def initialize(self) -> None:
        """Builds virtual envs etc."""
        self.sync_benchmarks_folder()

    def setup(self, *, force: bool):
        """Creates datasets for read benchmarks."""
        logger.info(f"setup_benchmarks for {self.ref} / {self.commit}")
        cmd = (
            f"pytest {PYTEST_OPTIONS} -s -nauto --benchmark-disable "
            f"-m setup_benchmarks --force-setup={force} "
            f"--where={self.where} "
            f"--icechunk-prefix=benchmarks/{self.prefix}/ "
            "benchmarks/"
        )
        logger.info(">>> " + cmd)
        self.execute(cmd, check=True)

    def run(self, *, pytest_extra: str = "") -> None:
        """Actually runs the benchmarks."""
        logger.info(f"running benchmarks for {self.ref} / {self.commit}")

        assert self.bench_store_dir is not None
        # Note: .benchmarks is the default location for pytest-benchmark
        cmd = (
            f"pytest {pytest_extra} "
            f"--benchmark-storage={self.bench_store_dir}/.benchmarks "
            f"--benchmark-save={self.where}_{self.clean_ref}_{self.commit} "
            f"--where={self.where} "
            f"--icechunk-prefix=benchmarks/{self.prefix}/ "
            f"{PYTEST_OPTIONS} "
            "benchmarks/"
        )
        print(cmd)

        self.execute(cmd, check=False)


class LocalRunner(Runner):
    activate: str = "source .venv/bin/activate"
    bench_store_dir = CURRENTDIR

    def __init__(self, *, ref: str, where: str, save_prefix: str):
        super().__init__(ref=ref, where=where, save_prefix="")
        suffix = self.commit
        self.base = f"{TMP}/icechunk-bench-{suffix}"
        self.cwd = f"{TMP}/icechunk-bench-{suffix}/icechunk"
        self.pycwd = f"{TMP}/icechunk-bench-{suffix}/icechunk/icechunk-python"

    def sync_benchmarks_folder(self):
        subprocess.run(["cp", "-r", "benchmarks", f"{self.pycwd}"], check=True)

    def execute(self, cmd: str, **kwargs) -> None:
        # don't stop if benchmarks fail
        subprocess.run(f"{self.activate} && {cmd}", cwd=self.pycwd, shell=True, **kwargs)

    def initialize(self) -> None:
        logger.info(f"Running initialize for {self.ref} in {self.base}")

        deps = get_benchmark_deps(f"{CURRENTDIR}/pyproject.toml")
        subprocess.run(["mkdir", "-p", self.pycwd], check=False)
        subprocess.run(["python3", "-m", "venv", ".venv"], cwd=self.pycwd, check=True)
        cmd = f"pip install {PIP_OPTIONS} {self.pip_github_url} {deps}"
        self.execute(cmd, check=True)
        super().initialize()

    def run(self, *, pytest_extra: str = "") -> None:
        super().run(pytest_extra=pytest_extra)


class CoiledRunner(Runner):
    bench_store_dir = "."

    def get_coiled_run_args(self) -> tuple[str]:
        ckwargs = self.get_coiled_kwargs()
        return (
            "coiled",
            "run",
            "--interactive",
            f"--name={ckwargs['name']}",
            f"--keepalive={ckwargs['keepalive']}",
            f"--workspace={ckwargs['workspace']}",  # cloud
            f"--vm-type={ckwargs['vm_type']}",
            f"--software={ckwargs['software']}",
            f"--region={ckwargs['region']}",
        )

    def get_coiled_kwargs(self):
        COILED_SOFTWARE = {
            "icechunk-v0.1.0-alpha.1": "icechunk-alpha-release",
            "icechunk-v0.1.0-alpha.12": "icechunk-alpha-12",
        }

        # using the default region here
        kwargs = get_coiled_kwargs(store=self.where)
        kwargs["software"] = COILED_SOFTWARE.get(
            self.ref, f"icechunk-bench-{self.commit}"
        )
        kwargs["name"] = f"icebench-{self.commit}-{self.where}"
        kwargs["keepalive"] = "10m"
        return kwargs

    def initialize(self) -> None:
        import coiled

        deps = get_benchmark_deps(f"{CURRENTDIR}/pyproject.toml").split(" ")

        ckwargs = self.get_coiled_kwargs()
        # repeated calls are a no-op!
        coiled.create_software_environment(
            name=ckwargs["software"],
            workspace=ckwargs["workspace"],
            conda={
                "channels": ["conda-forge"],
                "dependencies": ["rust", "python=3.12", "pip"],
            },
            pip=[self.pip_github_url, "coiled", *deps],
        )
        super().initialize()

    def execute(self, cmd, **kwargs):
        subprocess.run([*self.get_coiled_run_args(), cmd], **kwargs)

    def sync_benchmarks_folder(self) -> None:
        subprocess.run(
            [
                *self.get_coiled_run_args(),
                "--file",
                "benchmarks/",
                "ls -alh ./.benchmarks/",
            ],
            check=True,
        )

    def run(self, *, pytest_extra: str = "") -> None:
        super().run(pytest_extra=pytest_extra)
        filename = f"{self.save_prefix}/{self.where}_{self.clean_ref}_{self.commit}.json"
        # This is crappy; but for some reason coiled.function doesn't see the right files :/
        # So we need to use 'coiled run'; upload to a bucket; and then download from bucket
        # pytest-benchmark cannot write directly to a bucket yet, sadly.
        # TODO: explore a mounting a bucket as a volume.
        self.execute(f"sh benchmarks/most_recent.sh {filename}")


def read_latest_benchmark_json() -> str:
    files = sorted(
        glob.glob("./.benchmarks/*", recursive=True),
        key=os.path.getmtime,
        reverse=True,
    )
    with open(files[0]) as f:
        json = f.read()
    return json


def init_for_ref(runner: Runner) -> None:
    runner.initialize()


def run_there(where: str, *, args, save_prefix) -> None:
    if where == "local":
        runner_cls = LocalRunner
    else:
        runner_cls = CoiledRunner

    runners = tuple(
        runner_cls(ref=ref, where=where, save_prefix=save_prefix) for ref in args.refs
    )

    # we can only initialize in parallel since the two refs may have the same spec version.
    process_map(serial=args.serial, func=partial(init_for_ref), iterable=runners)

    if not args.skip_setup:
        for runner in runners:
            runner.setup(force=args.force_setup)

    # TODO: this could be parallelized for coiled runners
    for runner in tqdm.tqdm(runners):
        runner.run(pytest_extra=args.pytest)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("refs", help="refs to run benchmarks for", nargs="+")
    parser.add_argument("--pytest", help="passed to pytest", default="")
    parser.add_argument("--serial", action="store_true", default=False)
    parser.add_argument(
        "--where",
        help="where to run? [local|s3|s3_ob|gcs], combinations are allowed: [s3|gcs]",
        default="local",
    )
    parser.add_argument(
        "--skip-setup",
        help="skip setup step, useful for benchmarks that don't need data",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--force-setup", help="forced recreation of datasets?", type=bool, default=False
    )
    args = parser.parse_args()

    if "|" in args.where:
        where = args.where.split("|")
    else:
        where = (args.where,)

    save_prefix = rdms()

    process_map(
        serial=args.serial,
        func=partial(run_there, args=args, save_prefix=save_prefix),
        iterable=where,
    )

    if any(w != "local" for w in where):
        subprocess.run(
            [
                "aws",
                "s3",
                "cp",
                f"s3://earthmover-scratch/benchmarks/{save_prefix}/",
                f"/tmp/benchmarks/{save_prefix}/",
                "--recursive",
            ],
            check=True,
        )
    refs = args.refs

    # TODO: clean this up
    if where == ("local",) and len(refs) > 1:
        files = sorted(
            glob.glob("./.benchmarks/**/*.json", recursive=True),
            key=os.path.getmtime,
            reverse=True,
        )[: len(refs)]
    else:
        files = sorted(
            glob.glob(f"/tmp/benchmarks/{save_prefix}/*.json", recursive=True),
            key=os.path.getmtime,
            reverse=True,
        )
    #  TODO: Use `just` here when we figure that out.
    subprocess.run(
        [
            "pytest-benchmark",
            "compare",
            "--group=group,func",
            "--sort=fullname",
            "--columns=median",
            "--name=normal",
            *files,
        ]
    )


# Compare wish-list:
# 1. skip differences < X%
# 2. groupby
# 3. better names in summary table
# 5. Compare icechunk vs plain Zarr



from typing import Any

import pytest
from numpy.testing import assert_array_equal

from icechunk import IcechunkStore, Repository, in_memory_storage

pytest.importorskip("hypothesis")

import hypothesis.strategies as st
from hypothesis import assume, given, settings

from zarr.testing.strategies import arrays, numpy_arrays


def create() -> IcechunkStore:
    repo = Repository.create(in_memory_storage())
    return repo.writable_session("main").store


icechunk_stores = st.builds(create)


@settings(report_multiple_bugs=True, deadline=None)
@given(data=st.data(), nparray=numpy_arrays())
def test_roundtrip(data: st.DataObject, nparray: Any) -> None:
    # TODO: support size-0 arrays GH392
    assume(nparray.size > 0)

    zarray = data.draw(
        arrays(
            stores=icechunk_stores,
            arrays=st.just(nparray),
            zarr_formats=st.just(3),
        )
    )
    assert_array_equal(nparray, zarray[:])



import datetime
from collections.abc import AsyncIterator, Iterator
from contextlib import contextmanager
from typing import Any, Self, cast

from icechunk import ConflictSolver
from icechunk._icechunk_python import (
    Diff,
    GCSummary,
    PyRepository,
    RepositoryConfig,
    SnapshotInfo,
    Storage,
)
from icechunk.credentials import AnyCredential
from icechunk.session import Session
from icechunk.store import IcechunkStore


class Repository:
    """An Icechunk repository."""

    _repository: PyRepository

    def __init__(self, repository: PyRepository):
        self._repository = repository

    @classmethod
    def create(
        cls,
        storage: Storage,
        config: RepositoryConfig | None = None,
        authorize_virtual_chunk_access: dict[str, AnyCredential | None] | None = None,
    ) -> Self:
        """
        Create a new Icechunk repository.
        If one already exists at the given store location, an error will be raised.

        !!! warning
            Attempting to create a Repo concurrently in the same location from multiple processes is not safe.
            Instead, create a Repo once and then open it concurrently.

        Parameters
        ----------
        storage : Storage
            The storage configuration for the repository.
        config : RepositoryConfig, optional
            The repository configuration. If not provided, a default configuration will be used.
        authorize_virtual_chunk_access : dict[str, AnyCredential | None], optional
            Authorize Icechunk to access virtual chunks in these containers. A mapping
            from container url_prefix to the credentials to use to access chunks in
            that container. If credential is `None`, they will be fetched from the
            environment, or anonymous credentials will be used if the container allows it.
            As a security measure, Icechunk will block access to virtual chunks if the
            container is not authorized using this argument.

        Returns
        -------
        Self
            An instance of the Repository class.
        """
        return cls(
            PyRepository.create(
                storage,
                config=config,
                authorize_virtual_chunk_access=authorize_virtual_chunk_access,
            )
        )

    @classmethod
    async def create_async(
        cls,
        storage: Storage,
        config: RepositoryConfig | None = None,
        authorize_virtual_chunk_access: dict[str, AnyCredential | None] | None = None,
    ) -> Self:
        """
        Create a new Icechunk repository asynchronously.
        If one already exists at the given store location, an error will be raised.

        !!! warning
            Attempting to create a Repo concurrently in the same location from multiple processes is not safe.
            Instead, create a Repo once and then open it concurrently.

        Parameters
        ----------
        storage : Storage
            The storage configuration for the repository.
        config : RepositoryConfig, optional
            The repository configuration. If not provided, a default configuration will be used.
        authorize_virtual_chunk_access : dict[str, AnyCredential | None], optional
            Authorize Icechunk to access virtual chunks in these containers. A mapping
            from container url_prefix to the credentials to use to access chunks in
            that container. If credential is `None`, they will be fetched from the
            environment, or anonymous credentials will be used if the container allows it.
            As a security measure, Icechunk will block access to virtual chunks if the
            container is not authorized using this argument.

        Returns
        -------
        Self
            An instance of the Repository class.
        """
        return cls(
            await PyRepository.create_async(
                storage,
                config=config,
                authorize_virtual_chunk_access=authorize_virtual_chunk_access,
            )
        )

    @classmethod
    def open(
        cls,
        storage: Storage,
        config: RepositoryConfig | None = None,
        authorize_virtual_chunk_access: dict[str, AnyCredential | None] | None = None,
    ) -> Self:
        """
        Open an existing Icechunk repository.

        If no repository exists at the given storage location, an error will be raised.

        !!! warning
            This method must be used with care in a multiprocessing context.
            Read more in our [Parallel Write Guide](./parallel.md#uncooperative-distributed-writes).

        Parameters
        ----------
        storage : Storage
            The storage configuration for the repository.
        config : RepositoryConfig, optional
            The repository settings. If not provided, a default configuration will be
            loaded from the repository.
        authorize_virtual_chunk_access : dict[str, AnyCredential | None], optional
            Authorize Icechunk to access virtual chunks in these containers. A mapping
            from container url_prefix to the credentials to use to access chunks in
            that container. If credential is `None`, they will be fetched from the
            environment, or anonymous credentials will be used if the container allows it.
            As a security measure, Icechunk will block access to virtual chunks if the
            container is not authorized using this argument.

        Returns
        -------
        Self
            An instance of the Repository class.
        """
        return cls(
            PyRepository.open(
                storage,
                config=config,
                authorize_virtual_chunk_access=authorize_virtual_chunk_access,
            )
        )

    @classmethod
    async def open_async(
        cls,
        storage: Storage,
        config: RepositoryConfig | None = None,
        authorize_virtual_chunk_access: dict[str, AnyCredential | None] | None = None,
    ) -> Self:
        """
        Open an existing Icechunk repository asynchronously.

        If no repository exists at the given storage location, an error will be raised.

        !!! warning
            This method must be used with care in a multiprocessing context.
            Read more in our [Parallel Write Guide](./parallel.md#uncooperative-distributed-writes).

        Parameters
        ----------
        storage : Storage
            The storage configuration for the repository.
        config : RepositoryConfig, optional
            The repository settings. If not provided, a default configuration will be
            loaded from the repository.
        authorize_virtual_chunk_access : dict[str, AnyCredential | None], optional
            Authorize Icechunk to access virtual chunks in these containers. A mapping
            from container url_prefix to the credentials to use to access chunks in
            that container. If credential is `None`, they will be fetched from the
            environment, or anonymous credentials will be used if the container allows it.
            As a security measure, Icechunk will block access to virtual chunks if the
            container is not authorized using this argument.

        Returns
        -------
        Self
            An instance of the Repository class.
        """
        return cls(
            await PyRepository.open_async(
                storage,
                config=config,
                authorize_virtual_chunk_access=authorize_virtual_chunk_access,
            )
        )

    @classmethod
    def open_or_create(
        cls,
        storage: Storage,
        config: RepositoryConfig | None = None,
        authorize_virtual_chunk_access: dict[str, AnyCredential | None] | None = None,
    ) -> Self:
        """
        Open an existing Icechunk repository or create a new one if it does not exist.

        !!! warning
            This method must be used with care in a multiprocessing context.
            Read more in our [Parallel Write Guide](./parallel.md#uncooperative-distributed-writes).

            Attempting to create a Repo concurrently in the same location from multiple processes is not safe.
            Instead, create a Repo once and then open it concurrently.

        Parameters
        ----------
        storage : Storage
            The storage configuration for the repository.
        config : RepositoryConfig, optional
            The repository settings. If not provided, a default configuration will be
            loaded from the repository.
        authorize_virtual_chunk_access : dict[str, AnyCredential | None], optional
            Authorize Icechunk to access virtual chunks in these containers. A mapping
            from container url_prefix to the credentials to use to access chunks in
            that container. If credential is `None`, they will be fetched from the
            environment, or anonymous credentials will be used if the container allows it.
            As a security measure, Icechunk will block access to virtual chunks if the
            container is not authorized using this argument.

        Returns
        -------
        Self
            An instance of the Repository class.
        """
        return cls(
            PyRepository.open_or_create(
                storage,
                config=config,
                authorize_virtual_chunk_access=authorize_virtual_chunk_access,
            )
        )

    @classmethod
    async def open_or_create_async(
        cls,
        storage: Storage,
        config: RepositoryConfig | None = None,
        authorize_virtual_chunk_access: dict[str, AnyCredential | None] | None = None,
    ) -> Self:
        """
        Open an existing Icechunk repository or create a new one if it does not exist (async version).

        !!! warning
            This method must be used with care in a multiprocessing context.
            Read more in our [Parallel Write Guide](./parallel.md#uncooperative-distributed-writes).

            Attempting to create a Repo concurrently in the same location from multiple processes is not safe.
            Instead, create a Repo once and then open it concurrently.

        Parameters
        ----------
        storage : Storage
            The storage configuration for the repository.
        config : RepositoryConfig, optional
            The repository settings. If not provided, a default configuration will be
            loaded from the repository.
        authorize_virtual_chunk_access : dict[str, AnyCredential | None], optional
            Authorize Icechunk to access virtual chunks in these containers. A mapping
            from container url_prefix to the credentials to use to access chunks in
            that container. If credential is `None`, they will be fetched from the
            environment, or anonymous credentials will be used if the container allows it.
            As a security measure, Icechunk will block access to virtual chunks if the
            container is not authorized using this argument.

        Returns
        -------
        Self
            An instance of the Repository class.
        """
        return cls(
            await PyRepository.open_or_create_async(
                storage,
                config=config,
                authorize_virtual_chunk_access=authorize_virtual_chunk_access,
            )
        )

    @staticmethod
    def exists(storage: Storage) -> bool:
        """
        Check if a repository exists at the given storage location.

        Parameters
        ----------
        storage : Storage
            The storage configuration for the repository.

        Returns
        -------
        bool
            True if the repository exists, False otherwise.
        """
        return PyRepository.exists(storage)

    @staticmethod
    async def exists_async(storage: Storage) -> bool:
        """
        Check if a repository exists at the given storage location (async version).

        Parameters
        ----------
        storage : Storage
            The storage configuration for the repository.

        Returns
        -------
        bool
            True if the repository exists, False otherwise.
        """
        return await PyRepository.exists_async(storage)

    def __getstate__(self) -> object:
        return {
            "_repository": self._repository.as_bytes(),
        }

    def __setstate__(self, state: object) -> None:
        if not isinstance(state, dict):
            raise ValueError("Invalid repository state")
        self._repository = PyRepository.from_bytes(state["_repository"])

    @staticmethod
    def fetch_config(storage: Storage) -> RepositoryConfig | None:
        """
        Fetch the configuration for the repository saved in storage.

        Parameters
        ----------
        storage : Storage
            The storage configuration for the repository.

        Returns
        -------
        RepositoryConfig | None
            The repository configuration if it exists, None otherwise.
        """
        return PyRepository.fetch_config(storage)

    @staticmethod
    async def fetch_config_async(storage: Storage) -> RepositoryConfig | None:
        """
        Fetch the configuration for the repository saved in storage (async version).

        Parameters
        ----------
        storage : Storage
            The storage configuration for the repository.

        Returns
        -------
        RepositoryConfig | None
            The repository configuration if it exists, None otherwise.
        """
        return await PyRepository.fetch_config_async(storage)

    def save_config(self) -> None:
        """
        Save the repository configuration to storage, this configuration will be used in future calls to Repository.open.

        Returns
        -------
        None
        """
        return self._repository.save_config()

    async def save_config_async(self) -> None:
        """
        Save the repository configuration to storage (async version).

        Returns
        -------
        None
        """
        return await self._repository.save_config_async()

    @property
    def config(self) -> RepositoryConfig:
        """
        Get a copy of this repository's config.

        Returns
        -------
        RepositoryConfig
            The repository configuration.
        """
        return self._repository.config()

    @property
    def storage(self) -> Storage:
        """
        Get a copy of this repository's Storage instance.

        Returns
        -------
        Storage
            The repository storage instance.
        """
        return self._repository.storage()

    def reopen(
        self,
        config: RepositoryConfig | None = None,
        authorize_virtual_chunk_access: dict[str, AnyCredential | None] | None = None,
    ) -> Self:
        """
        Reopen the repository with new configuration or credentials.

        Parameters
        ----------
        config : RepositoryConfig, optional
            The new repository configuration. If not provided, uses the existing configuration.
        authorize_virtual_chunk_access : dict[str, AnyCredential | None], optional
            New virtual chunk access credentials.

        Returns
        -------
        Self
            A new Repository instance with the updated configuration.
        """
        return self.__class__(
            self._repository.reopen(
                config=config,
                authorize_virtual_chunk_access=authorize_virtual_chunk_access,
            )
        )

    async def reopen_async(
        self,
        config: RepositoryConfig | None = None,
        authorize_virtual_chunk_access: dict[str, AnyCredential | None] | None = None,
    ) -> Self:
        """
        Reopen the repository with new configuration or credentials (async version).

        Parameters
        ----------
        config : RepositoryConfig, optional
            The new repository configuration. If not provided, uses the existing configuration.
        authorize_virtual_chunk_access : dict[str, AnyCredential | None], optional
            New virtual chunk access credentials.

        Returns
        -------
        Self
            A new Repository instance with the updated configuration.
        """
        return self.__class__(
            await self._repository.reopen_async(
                config=config,
                authorize_virtual_chunk_access=authorize_virtual_chunk_access,
            )
        )

    def set_default_commit_metadata(self, metadata: dict[str, Any]) -> None:
        """
        Set the default commit metadata for the repository. This is useful for providing
        addition static system conexted metadata to all commits.

        When a commit is made, the metadata will be merged with the metadata provided, with any
        duplicate keys being overwritten by the metadata provided in the commit.

        !!! warning
            This metadata is only applied to sessions that are created after this call. Any open
            writable sessions will not be affected and will not use the new default metadata.

        Parameters
        ----------
        metadata : dict[str, Any]
            The default commit metadata. Pass an empty dict to clear the default metadata.
        """
        return self._repository.set_default_commit_metadata(metadata)

    def default_commit_metadata(self) -> dict[str, Any]:
        """
        Get the current configured default commit metadata for the repository.

        Returns
        -------
        dict[str, Any]
            The default commit metadata.
        """
        return self._repository.default_commit_metadata()

    def ancestry(
        self,
        *,
        branch: str | None = None,
        tag: str | None = None,
        snapshot_id: str | None = None,
    ) -> Iterator[SnapshotInfo]:
        """
        Get the ancestry of a snapshot.

        Parameters
        ----------
        branch : str, optional
            The branch to get the ancestry of.
        tag : str, optional
            The tag to get the ancestry of.
        snapshot_id : str, optional
            The snapshot ID to get the ancestry of.

        Returns
        -------
        list[SnapshotInfo]
            The ancestry of the snapshot, listing out the snapshots and their metadata.

        Notes
        -----
        Only one of the arguments can be specified.
        """

        # the returned object is both an Async and Sync iterator
        res = cast(
            Iterator[SnapshotInfo],
            self._repository.async_ancestry(
                branch=branch, tag=tag, snapshot_id=snapshot_id
            ),
        )
        return res

    def async_ancestry(
        self,
        *,
        branch: str | None = None,
        tag: str | None = None,
        snapshot_id: str | None = None,
    ) -> AsyncIterator[SnapshotInfo]:
        """
        Get the ancestry of a snapshot.

        Parameters
        ----------
        branch : str, optional
            The branch to get the ancestry of.
        tag : str, optional
            The tag to get the ancestry of.
        snapshot_id : str, optional
            The snapshot ID to get the ancestry of.

        Returns
        -------
        list[SnapshotInfo]
            The ancestry of the snapshot, listing out the snapshots and their metadata.

        Notes
        -----
        Only one of the arguments can be specified.
        """
        return self._repository.async_ancestry(
            branch=branch, tag=tag, snapshot_id=snapshot_id
        )

    def create_branch(self, branch: str, snapshot_id: str) -> None:
        """
        Create a new branch at the given snapshot.

        Parameters
        ----------
        branch : str
            The name of the branch to create.
        snapshot_id : str
            The snapshot ID to create the branch at.

        Returns
        -------
        None
        """
        self._repository.create_branch(branch, snapshot_id)

    async def create_branch_async(self, branch: str, snapshot_id: str) -> None:
        """
        Create a new branch at the given snapshot (async version).

        Parameters
        ----------
        branch : str
            The name of the branch to create.
        snapshot_id : str
            The snapshot ID to create the branch at.

        Returns
        -------
        None
        """
        await self._repository.create_branch_async(branch, snapshot_id)

    def list_branches(self) -> set[str]:
        """
        List the branches in the repository.

        Returns
        -------
        set[str]
            A set of branch names.
        """
        return self._repository.list_branches()

    async def list_branches_async(self) -> set[str]:
        """
        List the branches in the repository (async version).

        Returns
        -------
        set[str]
            A set of branch names.
        """
        return await self._repository.list_branches_async()

    def lookup_branch(self, branch: str) -> str:
        """
        Get the tip snapshot ID of a branch.

        Parameters
        ----------
        branch : str
            The branch to get the tip of.

        Returns
        -------
        str
            The snapshot ID of the tip of the branch.
        """
        return self._repository.lookup_branch(branch)

    async def lookup_branch_async(self, branch: str) -> str:
        """
        Get the tip snapshot ID of a branch (async version).

        Parameters
        ----------
        branch : str
            The branch to get the tip of.

        Returns
        -------
        str
            The snapshot ID of the tip of the branch.
        """
        return await self._repository.lookup_branch_async(branch)

    def lookup_snapshot(self, snapshot_id: str) -> SnapshotInfo:
        """
        Get the SnapshotInfo given a snapshot ID

        Parameters
        ----------
        snapshot_id : str
            The id of the snapshot to look up

        Returns
        -------
        SnapshotInfo
        """
        return self._repository.lookup_snapshot(snapshot_id)

    async def lookup_snapshot_async(self, snapshot_id: str) -> SnapshotInfo:
        """
        Get the SnapshotInfo given a snapshot ID (async version)

        Parameters
        ----------
        snapshot_id : str
            The id of the snapshot to look up

        Returns
        -------
        SnapshotInfo
        """
        return await self._repository.lookup_snapshot_async(snapshot_id)

    def reset_branch(self, branch: str, snapshot_id: str) -> None:
        """
        Reset a branch to a specific snapshot.

        This will permanently alter the history of the branch such that the tip of
        the branch is the specified snapshot.

        Parameters
        ----------
        branch : str
            The branch to reset.
        snapshot_id : str
            The snapshot ID to reset the branch to.

        Returns
        -------
        None
        """
        self._repository.reset_branch(branch, snapshot_id)

    async def reset_branch_async(self, branch: str, snapshot_id: str) -> None:
        """
        Reset a branch to a specific snapshot (async version).

        This will permanently alter the history of the branch such that the tip of
        the branch is the specified snapshot.

        Parameters
        ----------
        branch : str
            The branch to reset.
        snapshot_id : str
            The snapshot ID to reset the branch to.

        Returns
        -------
        None
        """
        await self._repository.reset_branch_async(branch, snapshot_id)

    def delete_branch(self, branch: str) -> None:
        """
        Delete a branch.

        Parameters
        ----------
        branch : str
            The branch to delete.

        Returns
        -------
        None
        """
        self._repository.delete_branch(branch)

    async def delete_branch_async(self, branch: str) -> None:
        """
        Delete a branch (async version).

        Parameters
        ----------
        branch : str
            The branch to delete.

        Returns
        -------
        None
        """
        await self._repository.delete_branch_async(branch)

    def delete_tag(self, tag: str) -> None:
        """
        Delete a tag.

        Parameters
        ----------
        tag : str
            The tag to delete.

        Returns
        -------
        None
        """
        self._repository.delete_tag(tag)

    async def delete_tag_async(self, tag: str) -> None:
        """
        Delete a tag (async version).

        Parameters
        ----------
        tag : str
            The tag to delete.

        Returns
        -------
        None
        """
        await self._repository.delete_tag_async(tag)

    def create_tag(self, tag: str, snapshot_id: str) -> None:
        """
        Create a new tag at the given snapshot.

        Parameters
        ----------
        tag : str
            The name of the tag to create.
        snapshot_id : str
            The snapshot ID to create the tag at.

        Returns
        -------
        None
        """
        self._repository.create_tag(tag, snapshot_id)

    async def create_tag_async(self, tag: str, snapshot_id: str) -> None:
        """
        Create a new tag at the given snapshot (async version).

        Parameters
        ----------
        tag : str
            The name of the tag to create.
        snapshot_id : str
            The snapshot ID to create the tag at.

        Returns
        -------
        None
        """
        await self._repository.create_tag_async(tag, snapshot_id)

    def list_tags(self) -> set[str]:
        """
        List the tags in the repository.

        Returns
        -------
        set[str]
            A set of tag names.
        """
        return self._repository.list_tags()

    async def list_tags_async(self) -> set[str]:
        """
        List the tags in the repository (async version).

        Returns
        -------
        set[str]
            A set of tag names.
        """
        return await self._repository.list_tags_async()

    def lookup_tag(self, tag: str) -> str:
        """
        Get the snapshot ID of a tag.

        Parameters
        ----------
        tag : str
            The tag to get the snapshot ID of.

        Returns
        -------
        str
            The snapshot ID of the tag.
        """
        return self._repository.lookup_tag(tag)

    async def lookup_tag_async(self, tag: str) -> str:
        """
        Get the snapshot ID of a tag (async version).

        Parameters
        ----------
        tag : str
            The tag to get the snapshot ID of.

        Returns
        -------
        str
            The snapshot ID of the tag.
        """
        return await self._repository.lookup_tag_async(tag)

    def diff(
        self,
        *,
        from_branch: str | None = None,
        from_tag: str | None = None,
        from_snapshot_id: str | None = None,
        to_branch: str | None = None,
        to_tag: str | None = None,
        to_snapshot_id: str | None = None,
    ) -> Diff:
        """
        Compute an overview of the operations executed from version `from` to version `to`.

        Both versions, `from` and `to`, must be identified. Identification can be done using a branch, tag or snapshot id.
        The styles used to identify the `from` and `to` versions can be different.

        The `from` version must be a member of the `ancestry` of `to`.

        Returns
        -------
        Diff
            The operations executed between the two versions
        """
        return self._repository.diff(
            from_branch=from_branch,
            from_tag=from_tag,
            from_snapshot_id=from_snapshot_id,
            to_branch=to_branch,
            to_tag=to_tag,
            to_snapshot_id=to_snapshot_id,
        )

    async def diff_async(
        self,
        *,
        from_branch: str | None = None,
        from_tag: str | None = None,
        from_snapshot_id: str | None = None,
        to_branch: str | None = None,
        to_tag: str | None = None,
        to_snapshot_id: str | None = None,
    ) -> Diff:
        """
        Compute an overview of the operations executed from version `from` to version `to` (async version).

        Both versions, `from` and `to`, must be identified. Identification can be done using a branch, tag or snapshot id.
        The styles used to identify the `from` and `to` versions can be different.

        The `from` version must be a member of the `ancestry` of `to`.

        Returns
        -------
        Diff
            The operations executed between the two versions
        """
        return await self._repository.diff_async(
            from_branch=from_branch,
            from_tag=from_tag,
            from_snapshot_id=from_snapshot_id,
            to_branch=to_branch,
            to_tag=to_tag,
            to_snapshot_id=to_snapshot_id,
        )

    def readonly_session(
        self,
        branch: str | None = None,
        *,
        tag: str | None = None,
        snapshot_id: str | None = None,
        as_of: datetime.datetime | None = None,
    ) -> Session:
        """
        Create a read-only session.

        This can be thought of as a read-only checkout of the repository at a given snapshot.
        When branch or tag are provided, the session will be based on the tip of the branch or
        the snapshot ID of the tag.

        Parameters
        ----------
        branch : str, optional
            If provided, the branch to create the session on.
        tag : str, optional
            If provided, the tag to create the session on.
        snapshot_id : str, optional
            If provided, the snapshot ID to create the session on.
        as_of: datetime.datetime, optional
            When combined with the branch argument, it will open the session at the last
            snapshot that is at or before this datetime

        Returns
        -------
        Session
            The read-only session, pointing to the specified snapshot, tag, or branch.

        Notes
        -----
        Only one of the arguments can be specified.
        """
        return Session(
            self._repository.readonly_session(
                branch=branch, tag=tag, snapshot_id=snapshot_id, as_of=as_of
            )
        )

    async def readonly_session_async(
        self,
        branch: str | None = None,
        *,
        tag: str | None = None,
        snapshot_id: str | None = None,
        as_of: datetime.datetime | None = None,
    ) -> Session:
        """
        Create a read-only session (async version).

        This can be thought of as a read-only checkout of the repository at a given snapshot.
        When branch or tag are provided, the session will be based on the tip of the branch or
        the snapshot ID of the tag.

        Parameters
        ----------
        branch : str, optional
            If provided, the branch to create the session on.
        tag : str, optional
            If provided, the tag to create the session on.
        snapshot_id : str, optional
            If provided, the snapshot ID to create the session on.
        as_of: datetime.datetime, optional
            When combined with the branch argument, it will open the session at the last
            snapshot that is at or before this datetime

        Returns
        -------
        Session
            The read-only session, pointing to the specified snapshot, tag, or branch.

        Notes
        -----
        Only one of the arguments can be specified.
        """
        return Session(
            await self._repository.readonly_session_async(
                branch=branch, tag=tag, snapshot_id=snapshot_id, as_of=as_of
            )
        )

    def writable_session(self, branch: str) -> Session:
        """
        Create a writable session on a branch.

        Like the read-only session, this can be thought of as a checkout of the repository at the
        tip of the branch. However, this session is writable and can be used to make changes to the
        repository. When ready, the changes can be committed to the branch, after which the session will
        become a read-only session on the new snapshot.

        Parameters
        ----------
        branch : str
            The branch to create the session on.

        Returns
        -------
        Session
            The writable session on the branch.
        """
        return Session(self._repository.writable_session(branch))

    async def writable_session_async(self, branch: str) -> Session:
        """
        Create a writable session on a branch (async version).

        Like the read-only session, this can be thought of as a checkout of the repository at the
        tip of the branch. However, this session is writable and can be used to make changes to the
        repository. When ready, the changes can be committed to the branch, after which the session will
        become a read-only session on the new snapshot.

        Parameters
        ----------
        branch : str
            The branch to create the session on.

        Returns
        -------
        Session
            The writable session on the branch.
        """
        return Session(await self._repository.writable_session_async(branch))

    @contextmanager
    def transaction(
        self,
        branch: str,
        *,
        message: str,
        metadata: dict[str, Any] | None = None,
        rebase_with: ConflictSolver | None = None,
        rebase_tries: int = 1_000,
    ) -> Iterator[IcechunkStore]:
        """
        Create a transaction on a branch.

        This is a context manager that creates a writable session on the specified branch.
        When the context is exited, the session will be committed to the branch
        using the specified message.

        Parameters
        ----------
        branch : str
            The branch to create the transaction on.
        message : str
            The commit message to use when committing the session.
        metadata : dict[str, Any] | None, optional
            Additional metadata to store with the commit snapshot.
        rebase_with : ConflictSolver | None, optional
            If other session committed while the current session was writing, use Session.rebase with this solver.
        rebase_tries : int, optional
            If other session committed while the current session was writing, use Session.rebase up to this many times in a loop.

        Yields
        -------
        store : IcechunkStore
            A Zarr Store which can be used to interact with the data in the repository.
        """
        session = self.writable_session(branch)
        yield session.store
        session.commit(
            message=message,
            metadata=metadata,
            rebase_with=rebase_with,
            rebase_tries=rebase_tries,
        )

    def expire_snapshots(
        self,
        older_than: datetime.datetime,
        *,
        delete_expired_branches: bool = False,
        delete_expired_tags: bool = False,
    ) -> set[str]:
        """Expire all snapshots older than a threshold.

        This processes snapshots found by navigating all references in
        the repo, tags first, branches leter, both in lexicographical order.

        Returns the ids of all snapshots considered expired and skipped
        from history. Notice that this snapshot are not necessarily
        available for garbage collection, they could still be pointed by
        ether refs.

        If `delete_expired_*` is set to True, branches or tags that, after the
        expiration process, point to expired snapshots directly, will be
        deleted.

        Danger
        ------
        This is an administrative operation, it should be run
        carefully. The repository can still operate concurrently while
        `expire_snapshots` runs, but other readers can get inconsistent
        views of the repository history.

        Parameters
        ----------
        older_than: datetime.datetime
            Expire snapshots older than this time.
        delete_expired_branches: bool, optional
            Whether to delete any branches that now have only expired snapshots.
        delete_expired_tags: bool, optional
            Whether to delete any tags associated with expired snapshots

        Returns
        -------
        set of expires snapshot IDs
        """
        return self._repository.expire_snapshots(
            older_than,
            delete_expired_branches=delete_expired_branches,
            delete_expired_tags=delete_expired_tags,
        )

    async def expire_snapshots_async(
        self,
        older_than: datetime.datetime,
        *,
        delete_expired_branches: bool = False,
        delete_expired_tags: bool = False,
    ) -> set[str]:
        """Expire all snapshots older than a threshold (async version).

        This processes snapshots found by navigating all references in
        the repo, tags first, branches leter, both in lexicographical order.

        Returns the ids of all snapshots considered expired and skipped
        from history. Notice that this snapshot are not necessarily
        available for garbage collection, they could still be pointed by
        ether refs.

        If `delete_expired_*` is set to True, branches or tags that, after the
        expiration process, point to expired snapshots directly, will be
        deleted.

        Danger
        ------
        This is an administrative operation, it should be run
        carefully. The repository can still operate concurrently while
        `expire_snapshots` runs, but other readers can get inconsistent
        views of the repository history.

        Parameters
        ----------
        older_than: datetime.datetime
            Expire snapshots older than this time.
        delete_expired_branches: bool, optional
            Whether to delete any branches that now have only expired snapshots.
        delete_expired_tags: bool, optional
            Whether to delete any tags associated with expired snapshots

        Returns
        -------
        set of expires snapshot IDs
        """
        return await self._repository.expire_snapshots_async(
            older_than,
            delete_expired_branches=delete_expired_branches,
            delete_expired_tags=delete_expired_tags,
        )

    def rewrite_manifests(
        self, message: str, *, branch: str, metadata: dict[str, Any] | None = None
    ) -> str:
        """
        Rewrite manifests for all arrays.

        This method will start a new writable session on the specified branch,
        rewrite manifests for all arrays, and then commits with the specifeid ``message``
        and ``metadata``.

        A JSON representation of the currently active splitting configuration will be
        stored in the commit's metadata under the key `"splitting_config"`.

        Parameters
        ----------
        message : str
            The message to write with the commit.
        branch: str
            The branch to commit to.
        metadata : dict[str, Any] | None, optional
            Additional metadata to store with the commit snapshot.

        Returns
        -------
        str
            The snapshot ID of the new commit.

        """
        return self._repository.rewrite_manifests(
            message, branch=branch, metadata=metadata
        )

    async def rewrite_manifests_async(
        self, message: str, *, branch: str, metadata: dict[str, Any] | None = None
    ) -> str:
        """
        Rewrite manifests for all arrays (async version).

        This method will start a new writable session on the specified branch,
        rewrite manifests for all arrays, and then commits with the specifeid ``message``
        and ``metadata``.

        A JSON representation of the currently active splitting configuration will be
        stored in the commit's metadata under the key `"splitting_config"`.

        Parameters
        ----------
        message : str
            The message to write with the commit.
        branch: str
            The branch to commit to.
        metadata : dict[str, Any] | None, optional
            Additional metadata to store with the commit snapshot.

        Returns
        -------
        str
            The snapshot ID of the new commit.

        """
        return await self._repository.rewrite_manifests_async(
            message, branch=branch, metadata=metadata
        )

    def garbage_collect(
        self,
        delete_object_older_than: datetime.datetime,
        *,
        dry_run: bool = False,
        max_snapshots_in_memory: int = 50,
        max_compressed_manifest_mem_bytes: int = 512 * 1024 * 1024,
        max_concurrent_manifest_fetches: int = 500,
    ) -> GCSummary:
        """Delete any objects no longer accessible from any branches or tags.

        Danger
        ------
        This is an administrative operation, it should be run
        carefully. The repository can still operate concurrently while
        `garbage_collect` runs, but other reades can get inconsistent
        views if they are trying to access the expired snapshots.

        Parameters
        ----------
        delete_object_older_than: datetime.datetime
            Delete objects older than this time.
        dry_run: bool : bool
            Report results but don't delete any objects
        max_snapshots_in_memory : int
            Don't prefetch more than this many Snapshots to memory.
        max_compressed_manifest_mem_bytes : int
            Don't use more than this memory to store compressed in-flight manifests.
        max_concurrent_manifest_fetches : int
            Don't run more than this many concurrent manifest fetches.

        Returns
        -------
        GCSummary
            Summary of objects deleted.
        """

        return self._repository.garbage_collect(
            delete_object_older_than,
            dry_run=dry_run,
            max_snapshots_in_memory=max_snapshots_in_memory,
            max_compressed_manifest_mem_bytes=max_compressed_manifest_mem_bytes,
            max_concurrent_manifest_fetches=max_concurrent_manifest_fetches,
        )

    async def garbage_collect_async(
        self,
        delete_object_older_than: datetime.datetime,
        *,
        dry_run: bool = False,
        max_snapshots_in_memory: int = 50,
        max_compressed_manifest_mem_bytes: int = 512 * 1024 * 1024,
        max_concurrent_manifest_fetches: int = 500,
    ) -> GCSummary:
        """Delete any objects no longer accessible from any branches or tags (async version).

        Danger
        ------
        This is an administrative operation, it should be run
        carefully. The repository can still operate concurrently while
        `garbage_collect` runs, but other reades can get inconsistent
        views if they are trying to access the expired snapshots.

        Parameters
        ----------
        delete_object_older_than: datetime.datetime
            Delete objects older than this time.
        dry_run: bool : bool
            Report results but don't delete any objects
        max_snapshots_in_memory : int
            Don't prefetch more than this many Snapshots to memory.
        max_compressed_manifest_mem_bytes : int
            Don't use more than this memory to store compressed in-flight manifests.
        max_concurrent_manifest_fetches : int
            Don't run more than this many concurrent manifest fetches.

        Returns
        -------
        GCSummary
            Summary of objects deleted.
        """

        return await self._repository.garbage_collect_async(
            delete_object_older_than,
            dry_run=dry_run,
            max_snapshots_in_memory=max_snapshots_in_memory,
            max_compressed_manifest_mem_bytes=max_compressed_manifest_mem_bytes,
            max_concurrent_manifest_fetches=max_concurrent_manifest_fetches,
        )

    def total_chunks_storage(
        self,
        *,
        max_snapshots_in_memory: int = 50,
        max_compressed_manifest_mem_bytes: int = 512 * 1024 * 1024,
        max_concurrent_manifest_fetches: int = 500,
    ) -> int:
        """Calculate the total storage used for chunks, in bytes .

        It reports the storage needed to store all snapshots in the repository that
        are reachable from any branches or tags. Unreachable snapshots can be generated
        by using `reset_branch` or `expire_snapshots`. The chunks for these snapshots
        are not included in the result, and they should probably be deleted using
        `garbage_collection`.

        The result includes only native chunks, not adding virtual or inline chunks.

        Parameters
        ----------
        max_snapshots_in_memory: int
            Don't prefetch more than this many Snapshots to memory.
        max_compressed_manifest_mem_bytes : int
            Don't use more than this memory to store compressed in-flight manifests.
        max_concurrent_manifest_fetches : int
            Don't run more than this many concurrent manifest fetches.
        """

        return self._repository.total_chunks_storage(
            max_snapshots_in_memory=max_snapshots_in_memory,
            max_compressed_manifest_mem_bytes=max_compressed_manifest_mem_bytes,
            max_concurrent_manifest_fetches=max_concurrent_manifest_fetches,
        )

    async def total_chunks_storage_async(
        self,
        *,
        max_snapshots_in_memory: int = 50,
        max_compressed_manifest_mem_bytes: int = 512 * 1024 * 1024,
        max_concurrent_manifest_fetches: int = 500,
    ) -> int:
        """Calculate the total storage used for chunks, in bytes (async version).

        It reports the storage needed to store all snapshots in the repository that
        are reachable from any branches or tags. Unreachable snapshots can be generated
        by using `reset_branch` or `expire_snapshots`. The chunks for these snapshots
        are not included in the result, and they should probably be deleted using
        `garbage_collection`.

        The result includes only native chunks, not adding virtual or inline chunks.

        Parameters
        ----------
        max_snapshots_in_memory: int
            Don't prefetch more than this many Snapshots to memory.
        max_compressed_manifest_mem_bytes : int
            Don't use more than this memory to store compressed in-flight manifests.
        max_concurrent_manifest_fetches : int
            Don't run more than this many concurrent manifest fetches.
        """

        return await self._repository.total_chunks_storage_async(
            max_snapshots_in_memory=max_snapshots_in_memory,
            max_compressed_manifest_mem_bytes=max_compressed_manifest_mem_bytes,
            max_concurrent_manifest_fetches=max_concurrent_manifest_fetches,
        )

    def inspect_snapshot(self, snapshot_id: str, *, pretty: bool = True) -> str:
        return self._repository.inspect_snapshot(snapshot_id, pretty=pretty)

    async def inspect_snapshot_async(
        self, snapshot_id: str, *, pretty: bool = True
    ) -> str:
        return await self._repository.inspect_snapshot_async(snapshot_id, pretty=pretty)



import os
import re
from unittest import mock

import icechunk as ic
import zarr


@mock.patch.dict(os.environ, {"ICECHUNK_LOG": "debug"}, clear=True)
def test_debug_logs_from_environment(capfd) -> None:
    ic.set_logs_filter(None)
    ic.Repository.create(storage=ic.in_memory_storage())
    assert "Creating Repository" in capfd.readouterr().out


@mock.patch.dict(os.environ, clear=True)
def test_no_logs_from_environment(capfd) -> None:
    ic.set_logs_filter(None)
    ic.Repository.create(storage=ic.in_memory_storage())
    assert capfd.readouterr().out == ""


@mock.patch.dict(os.environ, clear=True)
def test_change_log_levels_from_env(capfd) -> None:
    # first with logs disabled
    ic.set_logs_filter(None)
    ic.Repository.create(storage=ic.in_memory_storage())
    assert capfd.readouterr().out == ""

    # now with logs enabled
    with mock.patch.dict(os.environ, {"ICECHUNK_LOG": "debug"}, clear=True):
        ic.set_logs_filter(None)
        ic.Repository.create(storage=ic.in_memory_storage())
        assert "Creating Repository" in capfd.readouterr().out


def test_debug_logs_from_argument(capfd) -> None:
    ic.set_logs_filter("debug")
    ic.Repository.create(storage=ic.in_memory_storage())
    assert "Creating Repository" in capfd.readouterr().out


@mock.patch.dict(os.environ, {"ICECHUNK_LOG": "debug"}, clear=True)
def test_no_logs_from_argument(capfd) -> None:
    ic.set_logs_filter("false")
    ic.Repository.create(storage=ic.in_memory_storage())
    assert capfd.readouterr().out == ""


def test_change_log_levels_from_argument(capfd) -> None:
    # first with logs disabled
    ic.set_logs_filter("")
    ic.Repository.create(storage=ic.in_memory_storage())
    assert capfd.readouterr().out == ""

    # now with logs enabled
    ic.set_logs_filter("debug")
    ic.Repository.create(storage=ic.in_memory_storage())
    assert "Creating Repository" in capfd.readouterr().out


def test_warn_on_small_caches(capfd) -> None:
    # first with logs disabled
    ic.set_logs_filter("warn")
    config = ic.RepositoryConfig(
        caching=ic.CachingConfig(num_chunk_refs=0, num_snapshot_nodes=0),
    )
    repo = ic.Repository.create(storage=ic.in_memory_storage(), config=config)
    session = repo.writable_session("main")

    array1 = zarr.create_array(
        name="array1",
        store=session.store,
        shape=(10),
        dtype="int64",
        zarr_format=3,
        chunks=(2,),
        fill_value=-1,
    )
    array2 = zarr.create_array(
        name="array2",
        store=session.store,
        shape=(10),
        dtype="int64",
        zarr_format=3,
        chunks=(2,),
        fill_value=-1,
    )
    array1[:] = 42
    array2[:] = 42
    session.commit("msg")

    out = capfd.readouterr().out
    # we only warn once for manifests and once for snapshots
    assert len(re.findall("WARN", out)) == 2

    assert re.search("5 chunk references", out)
    assert re.search("keep 0 references", out)
    assert re.search("3 nodes", out)
    assert re.search("keep 0 nodes", out)


def test_misspelling_warnings_for_icechunk(capfd) -> None:
    """Test that misspelled variations of 'icechunk' trigger warnings."""
    # Test common misspellings
    ic.set_logs_filter("icehunk:trace")
    err_output = capfd.readouterr().err
    assert "WARNING" in err_output and "icehunk" in err_output

    ic.set_logs_filter("icecunk:debug")
    err_output = capfd.readouterr().err
    assert "WARNING" in err_output and "icecunk" in err_output

    ic.set_logs_filter("iecchunk:info")
    err_output = capfd.readouterr().err
    assert "WARNING" in err_output and "iecchunk" in err_output

    ic.set_logs_filter("ichunk:warn")
    err_output = capfd.readouterr().err
    assert "WARNING" in err_output and "ichunk" in err_output


def test_no_false_positives_for_correct_spelling(capfd) -> None:
    """Test that correct 'icechunk' spelling doesn't trigger warnings."""
    ic.set_logs_filter("icechunk:debug")
    assert "WARNING" not in capfd.readouterr().err

    ic.set_logs_filter("icechunk::repository:trace")
    assert "WARNING" not in capfd.readouterr().err


def test_no_warnings_for_generic_log_levels(capfd) -> None:
    """Test that generic log levels don't trigger warnings."""
    ic.set_logs_filter("debug")
    assert "WARNING" not in capfd.readouterr().err

    ic.set_logs_filter("trace")
    assert "WARNING" not in capfd.readouterr().err

    ic.set_logs_filter("info")
    assert "WARNING" not in capfd.readouterr().err


def test_no_warnings_for_unrelated_modules(capfd) -> None:
    """Test that unrelated module names don't trigger warnings."""
    ic.set_logs_filter("tokio:debug")
    assert "WARNING" not in capfd.readouterr().err

    ic.set_logs_filter("serde:info")
    assert "WARNING" not in capfd.readouterr().err

    ic.set_logs_filter("my_app::module:warn")
    assert "WARNING" not in capfd.readouterr().err


def test_complex_filter_strings_with_misspellings(capfd) -> None:
    """Test complex filter strings with multiple modules."""
    ic.set_logs_filter("icehunk:trace,debug,tokio::io:warn")
    err_output = capfd.readouterr().err
    assert "WARNING" in err_output and "icehunk" in err_output
    # Should not warn about tokio
    assert "tokio" not in err_output

    ic.set_logs_filter("debug,icecunk:info,warn")
    err_output = capfd.readouterr().err
    assert "WARNING" in err_output and "icecunk" in err_output


def test_nested_module_paths_with_misspellings(capfd) -> None:
    """Test nested module paths like icehunk::repository:trace."""
    ic.set_logs_filter("icehunk::repository:trace")
    err_output = capfd.readouterr().err
    assert "WARNING" in err_output and "icehunk" in err_output

    ic.set_logs_filter("icecunk::store::manager:debug")
    err_output = capfd.readouterr().err
    assert "WARNING" in err_output and "icecunk" in err_output


def test_short_strings_not_flagged(capfd) -> None:
    """Test that very short strings don't get flagged as misspellings."""
    ic.set_logs_filter("ice:debug")
    assert "WARNING" not in capfd.readouterr().err

    ic.set_logs_filter("chunk:info")
    assert "WARNING" not in capfd.readouterr().err



import uuid
from collections.abc import Sized
from datetime import UTC, datetime, timedelta
from pathlib import Path

import numpy as np
import pytest

import zarr
import zarr.core
import zarr.core.buffer
from icechunk import (
    IcechunkError,
    ObjectStoreConfig,
    RepositoryConfig,
    S3Options,
    VirtualChunkContainer,
    VirtualChunkSpec,
    containers_credentials,
    http_store,
    in_memory_storage,
    local_filesystem_storage,
    local_filesystem_store,
    s3_credentials,
    s3_store,
)
from icechunk.repository import Repository
from tests.conftest import write_chunks_to_minio


@pytest.mark.filterwarnings("ignore:datetime.datetime.utcnow")
@pytest.mark.parametrize("use_async", [True, False])
async def test_write_minio_virtual_refs(use_async) -> None:
    prefix = str(uuid.uuid4())
    etags = write_chunks_to_minio(
        [
            (f"{prefix}/chunk-1", b"first"),
            (f"{prefix}/chunk-2", b"second"),
        ],
    )

    config = RepositoryConfig.default()
    store_config = s3_store(
        region="us-east-1",
        endpoint_url="http://localhost:9000",
        allow_http=True,
        s3_compatible=True,
        force_path_style=True,
    )
    container = VirtualChunkContainer("s3://testbucket/", store_config)
    config.set_virtual_chunk_container(container)
    credentials = containers_credentials(
        {
            "s3://testbucket": s3_credentials(
                access_key_id="minio123", secret_access_key="minio123"
            )
        }
    )

    # Open the store
    repo = Repository.open_or_create(
        storage=in_memory_storage(),
        config=config,
        authorize_virtual_chunk_access=credentials,
    )
    session = repo.writable_session("main")
    store = session.store

    array = zarr.create_array(
        store, shape=(5, 1, 3), chunks=(1, 1, 1), dtype="i4", compressors=None
    )

    # We add the virtual chunk refs without checksum, with the right etag, and with the wrong wrong etag and datetime.
    # This way we can check retrieval operations that should fail
    old = datetime.now(UTC) - timedelta(weeks=1)
    new = datetime.now(UTC) + timedelta(minutes=1)

    chunks = [
        VirtualChunkSpec(
            index=[0, 0, 0],
            location=f"s3://testbucket/{prefix}/chunk-1",
            offset=0,
            length=4,
        ),
        VirtualChunkSpec(
            index=[1, 0, 0],
            location=f"s3://testbucket/{prefix}/chunk-1",
            offset=0,
            length=4,
            etag_checksum=etags[0],
        ),
        VirtualChunkSpec(
            index=[2, 0, 0],
            location=f"s3://testbucket/{prefix}/chunk-1",
            offset=0,
            length=4,
            etag_checksum="bad etag",
        ),
        VirtualChunkSpec(
            index=[3, 0, 0],
            location=f"s3://testbucket/{prefix}/chunk-1",
            offset=0,
            length=4,
            last_updated_at_checksum=old,
        ),
        VirtualChunkSpec(
            index=[4, 0, 0],
            location=f"s3://testbucket/{prefix}/chunk-1",
            offset=0,
            length=4,
            last_updated_at_checksum=new,
        ),
        VirtualChunkSpec(
            index=[0, 0, 1],
            location=f"s3://testbucket/{prefix}/chunk-2",
            offset=1,
            length=4,
        ),
        VirtualChunkSpec(
            index=[1, 0, 1],
            location=f"s3://testbucket/{prefix}/chunk-2",
            offset=1,
            length=4,
            etag_checksum=etags[1],
        ),
        VirtualChunkSpec(
            index=[2, 0, 1],
            location=f"s3://testbucket/{prefix}/chunk-2",
            offset=1,
            length=4,
            etag_checksum="bad etag",
        ),
        VirtualChunkSpec(
            index=[3, 0, 1],
            location=f"s3://testbucket/{prefix}/chunk-2",
            offset=1,
            length=4,
            last_updated_at_checksum=old,
        ),
        VirtualChunkSpec(
            index=[4, 0, 1],
            location=f"s3://testbucket/{prefix}/chunk-2",
            offset=1,
            length=4,
            last_updated_at_checksum=new,
        ),
        # we write a ref that simulates a lost chunk
        VirtualChunkSpec(
            index=[0, 0, 2],
            location=f"s3://testbucket/{prefix}/non-existing",
            offset=1,
            length=4,
        ),
        # we write one that doesn't pass container validation
        VirtualChunkSpec(
            index=[0, 0, 2],
            location=f"bad://testbucket/{prefix}/non-existing",
            offset=1,
            length=4,
        ),
    ]

    if use_async:
        res = await store.set_virtual_refs_async(
            array_path="/", validate_containers=True, chunks=chunks
        )
    else:
        res = store.set_virtual_refs(
            array_path="/", validate_containers=True, chunks=chunks
        )

    # we got the failed ref index
    assert res == [(0, 0, 2)]

    # can validate virtual chunk containers
    with pytest.raises(IcechunkError, match="invalid chunk location"):
        store.set_virtual_ref(
            "c/0/0/2",
            f"bad://testbucket/{prefix}/non-existing",
            offset=1,
            length=4,
            validate_container=True,
        )

    buffer_prototype = zarr.core.buffer.default_buffer_prototype()

    first = await store.get("c/0/0/0", prototype=buffer_prototype)
    assert first is not None
    assert first.to_bytes() == b"firs"  # codespell:ignore firs
    assert await store.get("c/1/0/0", prototype=buffer_prototype) == first
    assert await store.get("c/4/0/0", prototype=buffer_prototype) == first

    second = await store.get("c/0/0/1", prototype=buffer_prototype)
    assert second is not None
    assert second.to_bytes() == b"econ"
    assert await store.get("c/1/0/1", prototype=buffer_prototype) == second
    assert await store.get("c/4/0/1", prototype=buffer_prototype) == second

    assert array[0, 0, 0] == 1936877926
    assert array[0, 0, 1] == 1852793701

    # fetch uninitialized chunk should be None
    assert await store.get("c/0/0/3", prototype=buffer_prototype) is None

    # fetching a virtual ref that disappeared should be an exception
    with pytest.raises(IcechunkError):
        # TODO: we should include the key and other info in the exception
        await store.get("c/0/0/2", prototype=buffer_prototype)

    all_locations = set(session.all_virtual_chunk_locations())
    assert f"s3://testbucket/{prefix}/non-existing" in all_locations
    assert f"s3://testbucket/{prefix}/chunk-1" in all_locations
    assert f"s3://testbucket/{prefix}/chunk-2" in all_locations

    # check we cannot get refs with bad etag o that were modified

    with pytest.raises(IcechunkError, match="chunk has changed"):
        await store.get("c/2/0/0", prototype=buffer_prototype)
    with pytest.raises(IcechunkError, match="chunk has changed"):
        await store.get("c/3/0/0", prototype=buffer_prototype)
    with pytest.raises(IcechunkError, match="chunk has changed"):
        await store.get("c/2/0/1", prototype=buffer_prototype)
    with pytest.raises(IcechunkError, match="chunk has changed"):
        await store.get("c/3/0/1", prototype=buffer_prototype)

    _snapshot_id = session.commit("Add virtual refs")


@pytest.mark.parametrize(
    "container_type,url_prefix,store_config",
    [
        (
            "s3",
            "s3://earthmover-sample-data/",
            ObjectStoreConfig.S3(S3Options(region="us-east-1", anonymous=True)),
        ),
        (
            "http",
            "https://earthmover-sample-data.s3.amazonaws.com/",
            http_store(),
        ),
    ],
)
@pytest.mark.parametrize("use_async", [True, False])
async def test_public_virtual_refs(
    tmpdir: Path,
    container_type: str,
    url_prefix: str,
    store_config: ObjectStoreConfig.S3 | ObjectStoreConfig.Http,
    use_async: bool,
) -> None:
    config = RepositoryConfig.default()
    container = VirtualChunkContainer(url_prefix, store_config)
    config.set_virtual_chunk_container(container)

    repo = Repository.open_or_create(
        storage=local_filesystem_storage(f"{tmpdir}/virtual-{container_type}"),
        config=config,
        authorize_virtual_chunk_access={url_prefix: None},
    )
    session = repo.writable_session("main")
    store = session.store

    root = zarr.Group.from_store(store=store, zarr_format=3)
    year = root.require_array(
        name="year", shape=((72,)), chunks=((72,)), dtype="float32", compressors=None
    )

    file_path = f"{url_prefix}/netcdf/oscar_vel2018.nc"
    if use_async:
        await store.set_virtual_ref_async(
            "year/c/0",
            file_path,
            offset=22306,
            length=288,
        )
    else:
        store.set_virtual_ref(
            "year/c/0",
            file_path,
            offset=22306,
            length=288,
        )

    nodes = [n async for n in store.list()]
    assert "year/c/0" in nodes

    year_values = year[:]
    assert isinstance(year_values, Sized)
    assert len(year_values) == 72
    actual_values = np.array(
        [
            2018.0,
            2018.0139,
            2018.0278,
            2018.0416,
            2018.0555,
            2018.0695,
            2018.0834,
            2018.0972,
            2018.1111,
            2018.125,
            2018.1389,
            2018.1528,
            2018.1666,
            2018.1805,
            2018.1945,
            2018.2084,
            2018.2222,
            2018.2361,
            2018.25,
            2018.2639,
            2018.2778,
            2018.2916,
            2018.3055,
            2018.3195,
            2018.3334,
            2018.3472,
            2018.3611,
            2018.375,
            2018.3889,
            2018.4028,
            2018.4166,
            2018.4305,
            2018.4445,
            2018.4584,
            2018.4722,
            2018.4861,
            2018.5,
            2018.5139,
            2018.5278,
            2018.5416,
            2018.5555,
            2018.5695,
            2018.5834,
            2018.5972,
            2018.6111,
            2018.625,
            2018.6389,
            2018.6528,
            2018.6666,
            2018.6805,
            2018.6945,
            2018.7084,
            2018.7222,
            2018.7361,
            2018.75,
            2018.7639,
            2018.7778,
            2018.7916,
            2018.8055,
            2018.8195,
            2018.8334,
            2018.8472,
            2018.8611,
            2018.875,
            2018.8889,
            2018.9028,
            2018.9166,
            2018.9305,
            2018.9445,
            2018.9584,
            2018.9722,
            2018.9861,
        ],
        dtype="float32",
    )
    assert np.allclose(year_values, actual_values)


def test_error_on_nonexisting_virtual_chunk_container() -> None:
    repo = Repository.open_or_create(
        storage=in_memory_storage(),
    )
    session = repo.writable_session("main")
    store = session.store

    array = zarr.create_array(
        store, shape=(1,), chunks=(1,), dtype="i4", compressors=None
    )

    store.set_virtual_refs(
        array_path="/",
        validate_containers=False,
        chunks=[
            VirtualChunkSpec(
                index=[0],
                location="file:///foo",
                offset=0,
                length=4,
            ),
        ],
    )

    with pytest.raises(
        IcechunkError, match="file:///foo.* edit the repository configuration"
    ):
        array[0]


def test_error_on_non_authorized_virtual_chunk_container() -> None:
    store_config = local_filesystem_store("/foo")
    container = VirtualChunkContainer("file:///foo/", store_config)
    config = RepositoryConfig.default()
    config.set_virtual_chunk_container(container)
    repo = Repository.open_or_create(
        storage=in_memory_storage(),
        config=config,
    )

    session = repo.writable_session("main")
    store = session.store

    array = zarr.create_array(
        store, shape=(1,), chunks=(1,), dtype="i4", compressors=None
    )

    store.set_virtual_refs(
        array_path="/",
        validate_containers=False,
        chunks=[
            VirtualChunkSpec(
                index=[0],
                location="file:///foo/bar",
                offset=0,
                length=4,
            ),
        ],
    )

    with pytest.raises(IcechunkError, match="file:///foo.*authorize"):
        array[0]


def test_cannot_write_invalid_urls() -> None:
    repo = Repository.create(
        storage=in_memory_storage(),
    )
    session = repo.writable_session("main")
    store = session.store

    zarr.create_array(store, shape=(5,), chunks=(1,), dtype="i4", compressors=None)

    with pytest.raises(IcechunkError, match="error parsing"):
        store.set_virtual_ref(
            "c/0",
            "",
            offset=1,
            length=4,
        )

    with pytest.raises(IcechunkError, match="error parsing"):
        store.set_virtual_ref(
            "c/0",
            "some/relative/url",
            offset=1,
            length=4,
        )

    with pytest.raises(IcechunkError, match="error parsing"):
        store.set_virtual_ref(
            "c/0",
            "/some/url/without/protocol",
            offset=1,
            length=4,
        )

    with pytest.raises(IcechunkError, match="error parsing"):
        store.set_virtual_refs(
            array_path="/",
            chunks=[
                VirtualChunkSpec(
                    index=[0],
                    location="relative",
                    offset=0,
                    length=4,
                ),
            ],
        )



from collections.abc import Iterable
from typing import cast

import hypothesis.strategies as st

import icechunk as ic
import zarr
from zarr.core.metadata import ArrayV3Metadata


@st.composite
def splitting_configs(
    draw: st.DrawFn, *, arrays: Iterable[zarr.Array]
) -> ic.ManifestSplittingConfig:
    config_dict: dict[
        ic.ManifestSplitCondition,
        dict[
            ic.ManifestSplitDimCondition.Axis
            | ic.ManifestSplitDimCondition.DimensionName
            | ic.ManifestSplitDimCondition.Any,
            int,
        ],
    ] = {}
    for array in arrays:
        if draw(st.booleans()):
            array_condition = ic.ManifestSplitCondition.name_matches(
                array.path.split("/")[-1]
            )
        else:
            array_condition = ic.ManifestSplitCondition.path_matches(array.path)
        dimnames = (
            cast(ArrayV3Metadata, array.metadata).dimension_names or (None,) * array.ndim
        )
        dimsize_axis_names = draw(
            st.lists(
                st.sampled_from(
                    tuple(zip(array.shape, range(array.ndim), dimnames, strict=False))
                ),
                min_size=1,
                unique=True,
            )
        )
        for size, axis, dimname in dimsize_axis_names:
            if dimname is None or draw(st.booleans()):
                key = ic.ManifestSplitDimCondition.Axis(axis)
            else:
                key = ic.ManifestSplitDimCondition.DimensionName(dimname)  # type: ignore[assignment]
            config_dict[array_condition] = {
                key: draw(st.integers(min_value=1, max_value=size + 10))
            }
    return ic.ManifestSplittingConfig.from_dict(config_dict)



import asyncio
from datetime import timedelta
from typing import cast

import pytest

import icechunk as ic
import zarr
import zarr.core
import zarr.core.array
import zarr.core.buffer


async def async_ancestry(
    repo: ic.Repository, **kwargs: str | None
) -> list[ic.SnapshotInfo]:
    return [parent async for parent in repo.async_ancestry(**kwargs)]


def test_timetravel() -> None:
    config = ic.RepositoryConfig.default()
    config.inline_chunk_threshold_bytes = 1
    repo = ic.Repository.create(
        storage=ic.in_memory_storage(),
        config=config,
    )

    session = repo.writable_session("main")
    store = session.store

    group = zarr.group(store=store, overwrite=True)
    air_temp = group.create_array(
        "air_temp", shape=(1000, 1000), chunks=(100, 100), dtype="i4"
    )

    air_temp[:, :] = 42
    assert air_temp[200, 6] == 42

    status = session.status()
    assert status.new_groups == {"/"}
    assert status.new_arrays == {"/air_temp"}
    assert list(status.updated_chunks.keys()) == ["/air_temp"]
    assert sorted(status.updated_chunks["/air_temp"]) == sorted(
        [[i, j] for i in range(10) for j in range(10)]
    )
    assert status.deleted_groups == set()
    assert status.deleted_arrays == set()
    assert status.updated_arrays == set()
    assert status.updated_groups == set()

    first_snapshot_id = session.commit("commit 1")
    assert session.read_only

    session = repo.writable_session("main")
    store = session.store
    group = zarr.open_group(store=store)
    air_temp = cast(zarr.core.array.Array, group["air_temp"])

    air_temp[:, :] = 54
    assert air_temp[200, 6] == 54

    new_snapshot_id = session.commit("commit 2")

    session = repo.readonly_session(snapshot_id=first_snapshot_id)
    store = session.store
    group = zarr.open_group(store=store, mode="r")
    air_temp = cast(zarr.core.array.Array, group["air_temp"])
    assert store.read_only
    assert air_temp[200, 6] == 42

    session = repo.readonly_session(snapshot_id=new_snapshot_id)
    store = session.store
    group = zarr.open_group(store=store, mode="r")
    air_temp = cast(zarr.core.array.Array, group["air_temp"])
    assert store.read_only
    assert air_temp[200, 6] == 54

    session = repo.writable_session("main")
    store = session.store
    group = zarr.open_group(store=store)
    air_temp = cast(zarr.core.array.Array, group["air_temp"])

    air_temp[:, :] = 76
    assert session.has_uncommitted_changes
    assert session.branch == "main"
    assert session.snapshot_id == new_snapshot_id

    session.discard_changes()
    assert not (session.has_uncommitted_changes)
    # I don't understand why I need to ignore here
    assert air_temp[200, 6] == 54  # type: ignore [unreachable]

    repo.create_branch("feature", new_snapshot_id)
    session = repo.writable_session("feature")
    store = session.store
    assert not store._read_only
    assert session.branch == "feature"

    group = zarr.open_group(store=store)
    air_temp = cast(zarr.core.array.Array, group["air_temp"])
    air_temp[:, :] = 90
    feature_snapshot_id = session.commit("commit 3")

    branches = repo.list_branches()
    assert branches == set(["main", "feature"])

    repo.delete_branch("feature")
    branches = repo.list_branches()
    assert branches == set(["main"])

    repo.create_tag("v1.0", feature_snapshot_id)
    repo.create_branch("feature-not-dead", feature_snapshot_id)
    session = repo.readonly_session(tag="v1.0")
    store = session.store
    assert store._read_only
    assert session.branch is None

    group = zarr.open_group(store=store, mode="r")
    air_temp = cast(zarr.core.array.Array, group["air_temp"])
    assert air_temp[200, 6] == 90

    parents = list(repo.ancestry(snapshot_id=feature_snapshot_id))
    assert [snap.message for snap in parents] == [
        "commit 3",
        "commit 2",
        "commit 1",
        "Repository initialized",
    ]
    assert parents[-1].id == "1CECHNKREP0F1RSTCMT0"
    assert [len(snap.manifests) for snap in parents] == [1, 1, 1, 0]
    assert sorted(parents, key=lambda p: p.written_at) == list(reversed(parents))
    assert len(set([snap.id for snap in parents])) == 4
    assert list(repo.ancestry(tag="v1.0")) == parents
    assert list(repo.ancestry(branch="feature-not-dead")) == parents

    diff = repo.diff(to_tag="v1.0", from_snapshot_id=parents[-1].id)
    assert diff.new_groups == {"/"}
    assert diff.new_arrays == {"/air_temp"}
    assert list(diff.updated_chunks.keys()) == ["/air_temp"]
    assert sorted(diff.updated_chunks["/air_temp"]) == sorted(
        [[i, j] for i in range(10) for j in range(10)]
    )
    assert diff.deleted_groups == set()
    assert diff.deleted_arrays == set()
    assert diff.updated_arrays == set()
    assert diff.updated_groups == set()
    assert (
        repr(diff)
        == """\
Groups created:
    /

Arrays created:
    /air_temp

Chunks updated:
    /air_temp:
        [0, 0]
        [0, 1]
        [0, 2]
        [0, 3]
        [0, 4]
        [0, 5]
        [0, 6]
        [0, 7]
        [0, 8]
        [0, 9]
        ... 90 more
"""
    )

    session = repo.writable_session("main")
    store = session.store

    group = zarr.open_group(store=store)
    air_temp = group.create_array(
        "air_temp", shape=(1000, 1000), chunks=(100, 100), dtype="i4", overwrite=True
    )
    assert (
        repr(session.status())
        == """\
Arrays created:
    /air_temp

Arrays deleted:
    /air_temp

"""
    )

    with pytest.raises(ic.IcechunkError, match="doesn't include"):
        # if we call diff in the wrong order it fails with a message
        repo.diff(from_tag="v1.0", to_snapshot_id=parents[-1].id)

    # check async ancestry works
    assert list(repo.ancestry(snapshot_id=feature_snapshot_id)) == asyncio.run(
        async_ancestry(repo, snapshot_id=feature_snapshot_id)
    )
    assert list(repo.ancestry(tag="v1.0")) == asyncio.run(
        async_ancestry(repo, tag="v1.0")
    )
    assert list(repo.ancestry(branch="feature-not-dead")) == asyncio.run(
        async_ancestry(repo, branch="feature-not-dead")
    )

    tags = repo.list_tags()
    assert tags == set(["v1.0"])
    tag_snapshot_id = repo.lookup_tag("v1.0")
    assert tag_snapshot_id == feature_snapshot_id

    actual = next(iter(repo.ancestry(tag="v1.0")))
    assert actual == repo.lookup_snapshot(actual.id)


async def test_branch_reset() -> None:
    config = ic.RepositoryConfig.default()
    config.inline_chunk_threshold_bytes = 1
    repo = ic.Repository.create(
        storage=ic.in_memory_storage(),
        config=config,
    )

    session = repo.writable_session("main")
    store = session.store

    group = zarr.group(store=store, overwrite=True)
    group.create_group("a")
    prev_snapshot_id = session.commit("group a")

    session = repo.writable_session("main")
    store = session.store

    group = zarr.open_group(store=store)
    group.create_group("b")
    session.commit("group b")

    keys = {k async for k in store.list()}
    assert "a/zarr.json" in keys
    assert "b/zarr.json" in keys

    repo.reset_branch("main", prev_snapshot_id)

    session = repo.readonly_session("main")
    store = session.store

    keys = {k async for k in store.list()}
    assert "a/zarr.json" in keys
    assert "b/zarr.json" not in keys

    assert (
        await store.get("b/zarr.json", zarr.core.buffer.default_buffer_prototype())
    ) is None


async def test_tag_delete() -> None:
    repo = ic.Repository.create(
        storage=ic.in_memory_storage(),
    )

    snap = repo.lookup_branch("main")
    repo.create_tag("tag", snap)
    repo.delete_tag("tag")

    with pytest.raises(ic.IcechunkError):
        repo.delete_tag("tag")

    with pytest.raises(ic.IcechunkError):
        repo.create_tag("tag", snap)


async def test_session_with_as_of() -> None:
    repo = ic.Repository.create(
        storage=ic.in_memory_storage(),
    )

    session = repo.writable_session("main")
    store = session.store

    times = []
    group = zarr.group(store=store, overwrite=True)
    sid = session.commit("root")
    times.append(next(repo.ancestry(snapshot_id=sid)).written_at)

    for i in range(5):
        session = repo.writable_session("main")
        store = session.store
        group = zarr.open_group(store=store)
        group.create_group(f"child {i}")
        sid = session.commit(f"child {i}")
        times.append(next(repo.ancestry(snapshot_id=sid)).written_at)

    ancestry = list(p for p in repo.ancestry(branch="main"))
    assert len(ancestry) == 7  # initial + root + 5 children

    store = repo.readonly_session("main", as_of=times[-1]).store
    group = zarr.open_group(store=store, mode="r")

    for i, time in enumerate(times):
        store = repo.readonly_session("main", as_of=time).store
        group = zarr.open_group(store=store, mode="r")
        expected_children = {f"child {j}" for j in range(i)}
        actual_children = {g[0] for g in group.members()}
        assert expected_children == actual_children


async def test_default_commit_metadata() -> None:
    repo = ic.Repository.create(
        storage=ic.in_memory_storage(),
    )

    repo.set_default_commit_metadata({"user": "test"})
    session = repo.writable_session("main")
    root = zarr.group(store=session.store, overwrite=True)
    root.create_group("child")
    sid = session.commit("root")
    snap = next(repo.ancestry(snapshot_id=sid))
    assert snap.metadata == {"user": "test"}


async def test_timetravel_async() -> None:
    config = ic.RepositoryConfig.default()
    config.inline_chunk_threshold_bytes = 1
    repo = await ic.Repository.create_async(
        storage=ic.in_memory_storage(),
        config=config,
    )

    session = await repo.writable_session_async("main")
    group = zarr.group(store=session.store, overwrite=True)
    air_temp = group.create_array(
        "air_temp", shape=(1000, 1000), chunks=(100, 100), dtype="i4"
    )

    air_temp[:, :] = 42
    assert air_temp[200, 6] == 42

    status = session.status()
    assert status.new_groups == {"/"}
    assert status.new_arrays == {"/air_temp"}
    assert list(status.updated_chunks.keys()) == ["/air_temp"]
    assert sorted(status.updated_chunks["/air_temp"]) == sorted(
        [[i, j] for i in range(10) for j in range(10)]
    )
    assert status.deleted_groups == set()
    assert status.deleted_arrays == set()
    assert status.updated_arrays == set()
    assert status.updated_groups == set()

    first_snapshot_id = await session.commit_async("commit 1")
    assert session.read_only

    session = await repo.writable_session_async("main")
    group = zarr.open_group(store=session.store)
    air_temp = cast(zarr.core.array.Array, group["air_temp"])

    air_temp[:, :] = 54
    assert air_temp[200, 6] == 54

    new_snapshot_id = await session.commit_async("commit 2")

    session = await repo.readonly_session_async(snapshot_id=first_snapshot_id)
    store = session.store
    group = zarr.open_group(store=store, mode="r")
    air_temp = cast(zarr.core.array.Array, group["air_temp"])
    assert store.read_only
    assert air_temp[200, 6] == 42

    session = await repo.readonly_session_async(snapshot_id=new_snapshot_id)
    store = session.store
    group = zarr.open_group(store=store, mode="r")
    air_temp = cast(zarr.core.array.Array, group["air_temp"])
    assert store.read_only
    assert air_temp[200, 6] == 54

    session = await repo.writable_session_async("main")
    group = zarr.open_group(store=session.store)
    air_temp = cast(zarr.core.array.Array, group["air_temp"])

    air_temp[:, :] = 76
    assert session.has_uncommitted_changes
    assert session.branch == "main"
    assert session.snapshot_id == new_snapshot_id

    session.discard_changes()
    assert not session.has_uncommitted_changes
    assert air_temp[200, 6] == 54

    await repo.create_branch_async("feature", new_snapshot_id)
    session = await repo.writable_session_async("feature")
    assert not session.store.read_only
    assert session.branch == "feature"

    group = zarr.open_group(store=session.store)
    air_temp = cast(zarr.core.array.Array, group["air_temp"])
    air_temp[:, :] = 90
    feature_snapshot_id = await session.commit_async("commit 3")

    branches = await repo.list_branches_async()
    assert branches == set(["main", "feature"])

    await repo.delete_branch_async("feature")
    branches = await repo.list_branches_async()
    assert branches == set(["main"])

    await repo.create_tag_async("v1.0", feature_snapshot_id)
    await repo.create_branch_async("feature-not-dead", feature_snapshot_id)
    session = await repo.readonly_session_async(tag="v1.0")
    assert session.store.read_only
    assert session.branch is None

    group = zarr.open_group(store=session.store, mode="r")
    air_temp = cast(zarr.core.array.Array, group["air_temp"])
    assert air_temp[200, 6] == 90

    parents = [
        parent async for parent in repo.async_ancestry(snapshot_id=feature_snapshot_id)
    ]
    assert [snap.message for snap in parents] == [
        "commit 3",
        "commit 2",
        "commit 1",
        "Repository initialized",
    ]
    assert parents[-1].id == "1CECHNKREP0F1RSTCMT0"
    assert [len(snap.manifests) for snap in parents] == [1, 1, 1, 0]
    assert sorted(parents, key=lambda p: p.written_at) == list(reversed(parents))
    assert len(set([snap.id for snap in parents])) == 4
    assert [parent async for parent in repo.async_ancestry(tag="v1.0")] == parents
    assert [
        parent async for parent in repo.async_ancestry(branch="feature-not-dead")
    ] == parents

    diff = await repo.diff_async(to_tag="v1.0", from_snapshot_id=parents[-1].id)
    assert diff.new_groups == {"/"}
    assert diff.new_arrays == {"/air_temp"}
    assert list(diff.updated_chunks.keys()) == ["/air_temp"]
    assert sorted(diff.updated_chunks["/air_temp"]) == sorted(
        [[i, j] for i in range(10) for j in range(10)]
    )
    assert diff.deleted_groups == set()
    assert diff.deleted_arrays == set()
    assert diff.updated_arrays == set()
    assert diff.updated_groups == set()
    assert (
        repr(diff)
        == """\
Groups created:
    /

Arrays created:
    /air_temp

Chunks updated:
    /air_temp:
        [0, 0]
        [0, 1]
        [0, 2]
        [0, 3]
        [0, 4]
        [0, 5]
        [0, 6]
        [0, 7]
        [0, 8]
        [0, 9]
        ... 90 more
"""
    )

    session = await repo.writable_session_async("main")
    group = zarr.open_group(store=session.store)
    air_temp = group.create_array(
        "air_temp", shape=(1000, 1000), chunks=(100, 100), dtype="i4", overwrite=True
    )
    assert (
        repr(session.status())
        == """\
Arrays created:
    /air_temp

Arrays deleted:
    /air_temp

"""
    )

    with pytest.raises(ic.IcechunkError, match="doesn't include"):
        # if we call diff in the wrong order it fails with a message
        await repo.diff_async(from_tag="v1.0", to_snapshot_id=parents[-1].id)

    tags = await repo.list_tags_async()
    assert tags == set(["v1.0"])
    tag_snapshot_id = await repo.lookup_tag_async("v1.0")
    assert tag_snapshot_id == feature_snapshot_id

    actual = next(iter([parent async for parent in repo.async_ancestry(tag="v1.0")]))
    assert actual == await repo.lookup_snapshot_async(actual.id)


async def test_branch_reset_async() -> None:
    config = ic.RepositoryConfig.default()
    config.inline_chunk_threshold_bytes = 1
    repo = await ic.Repository.create_async(
        storage=ic.in_memory_storage(),
        config=config,
    )

    session = await repo.writable_session_async("main")
    group = zarr.group(store=session.store, overwrite=True)
    group.create_group("a")
    prev_snapshot_id = await session.commit_async("group a")

    session = await repo.writable_session_async("main")
    store = session.store
    group = zarr.open_group(store=store)
    group.create_group("b")
    await session.commit_async("group b")

    keys = {k async for k in store.list()}
    assert "a/zarr.json" in keys
    assert "b/zarr.json" in keys

    await repo.reset_branch_async("main", prev_snapshot_id)

    session = await repo.readonly_session_async("main")
    store = session.store
    keys = {k async for k in store.list()}
    assert "a/zarr.json" in keys
    assert "b/zarr.json" not in keys

    assert (
        await store.get("b/zarr.json", zarr.core.buffer.default_buffer_prototype())
    ) is None


async def test_tag_delete_async() -> None:
    repo = await ic.Repository.create_async(
        storage=ic.in_memory_storage(),
    )

    snap = await repo.lookup_branch_async("main")
    await repo.create_tag_async("tag", snap)
    await repo.delete_tag_async("tag")

    with pytest.raises(ic.IcechunkError):
        await repo.delete_tag_async("tag")

    with pytest.raises(ic.IcechunkError):
        await repo.create_tag_async("tag", snap)


async def test_session_with_as_of_async() -> None:
    repo = await ic.Repository.create_async(
        storage=ic.in_memory_storage(),
    )

    session = await repo.writable_session_async("main")

    times = []
    group = zarr.group(store=session.store, overwrite=True)
    sid = await session.commit_async("root")
    to_append = await anext(repo.async_ancestry(snapshot_id=sid))
    times.append(to_append.written_at)

    for i in range(5):
        session = await repo.writable_session_async("main")
        group = zarr.open_group(store=session.store)
        group.create_group(f"child {i}")
        sid = await session.commit_async(f"child {i}")
        to_append = await anext(repo.async_ancestry(snapshot_id=sid))
        times.append(to_append.written_at)

    ancestry = [p async for p in repo.async_ancestry(branch="main")]
    assert len(ancestry) == 7  # initial + root + 5 children

    store = (await repo.readonly_session_async("main", as_of=times[-1])).store
    group = zarr.open_group(store=store, mode="r")

    for i, time in enumerate(times):
        store = (await repo.readonly_session_async("main", as_of=time)).store
        group = zarr.open_group(store=store, mode="r")
        expected_children = {f"child {j}" for j in range(i)}
        actual_children = {g[0] for g in group.members()}
        assert expected_children == actual_children


async def test_tag_expiration_async() -> None:
    repo = await ic.Repository.create_async(storage=ic.in_memory_storage())
    session = await repo.writable_session_async("main")
    root = zarr.group(store=session.store, overwrite=True)
    a = await session.commit_async("a")

    session = await repo.writable_session_async("main")
    root = zarr.group(store=session.store, overwrite=True)
    root.create_group("child1")
    b = await session.commit_async("b")

    session = await repo.writable_session_async("main")
    root = zarr.group(store=session.store, overwrite=True)
    root.create_group("child2")
    await session.commit_async("c")

    await repo.create_tag_async("0", a)
    await repo.create_tag_async("1", b)
    assert await repo.list_tags_async() == {"0", "1"}

    expired = await repo.expire_snapshots_async(
        (await repo.lookup_snapshot_async(b)).written_at + timedelta(microseconds=1),
        delete_expired_tags=True,
    )
    assert expired == {a, b}
    assert await repo.list_tags_async() == set()


async def test_branch_expiration_async() -> None:
    repo = await ic.Repository.open_or_create_async(storage=ic.in_memory_storage())
    session = await repo.writable_session_async("main")
    root = zarr.group(store=session.store, overwrite=True)
    a = await session.commit_async("a")

    await repo.create_branch_async("branch", a)
    session = await repo.writable_session_async("branch")
    root = zarr.group(store=session.store, overwrite=True)
    root.create_group("child1")
    b = await session.commit_async("b")

    session = await repo.writable_session_async("main")
    root = zarr.group(store=session.store, overwrite=True)
    root.create_group("child2")
    c = await session.commit_async("c")

    expired = await repo.expire_snapshots_async(
        (await repo.lookup_snapshot_async(b)).written_at + timedelta(microseconds=1),
        delete_expired_branches=True,
    )
    assert expired == {a, b}
    assert "branch" not in await repo.list_branches_async()

    for snap in (a, b):
        await repo.lookup_snapshot_async(snap)

    # FIXME: this fails to delete snapshot `b` with microseconds=1
    await repo.garbage_collect_async(
        (await repo.lookup_snapshot_async(b)).written_at + timedelta(seconds=1)
    )
    # make sure snapshot cannot be opened anymore
    for snap in (a, b):
        with pytest.raises(ic.IcechunkError):
            await repo.readonly_session_async(snapshot_id=snap)

    # should succeed
    await repo.lookup_snapshot_async(c)


def test_tag_expiration() -> None:
    repo = ic.Repository.create(storage=ic.in_memory_storage())
    session = repo.writable_session("main")
    root = zarr.group(store=session.store, overwrite=True)
    a = session.commit("a")

    session = repo.writable_session("main")
    root = zarr.group(store=session.store, overwrite=True)
    root.create_group("child1")
    b = session.commit("b")

    session = repo.writable_session("main")
    root = zarr.group(store=session.store, overwrite=True)
    root.create_group("child2")
    session.commit("c")

    repo.create_tag("0", a)
    repo.create_tag("1", b)
    assert repo.list_tags() == {"0", "1"}

    expired = repo.expire_snapshots(
        repo.lookup_snapshot(b).written_at + timedelta(microseconds=1),
        delete_expired_tags=True,
    )
    assert expired == {a, b}
    assert repo.list_tags() == set()


def test_branch_expiration() -> None:
    repo = ic.Repository.create(storage=ic.in_memory_storage())
    session = repo.writable_session("main")
    root = zarr.group(store=session.store, overwrite=True)
    a = session.commit("a")

    repo.create_branch("branch", a)
    session = repo.writable_session("branch")
    root = zarr.group(store=session.store, overwrite=True)
    root.create_group("child1")
    b = session.commit("b")

    session = repo.writable_session("main")
    root = zarr.group(store=session.store, overwrite=True)
    root.create_group("child2")
    c = session.commit("c")

    expired = repo.expire_snapshots(
        repo.lookup_snapshot(b).written_at + timedelta(microseconds=1),
        delete_expired_branches=True,
    )
    assert expired == {a, b}
    assert "branch" not in repo.list_branches()

    for snap in (a, b):
        repo.lookup_snapshot(snap)

    # FIXME: this fails to delete snapshot `b` with microseconds=1
    repo.garbage_collect(repo.lookup_snapshot(b).written_at + timedelta(seconds=1))
    # make sure snapshot cannot be opened anymore
    for snap in (a, b):
        with pytest.raises(ic.IcechunkError):
            repo.readonly_session(snapshot_id=snap)

    # should succeed
    repo.lookup_snapshot(c)


async def test_repository_lifecycle_async() -> None:
    """Test Repository configuration and lifecycle async methods."""
    storage = ic.in_memory_storage()

    # Test exists_async with non-existent repo
    assert not await ic.Repository.exists_async(storage)

    # Test fetch_config_async with non-existent repo
    config = await ic.Repository.fetch_config_async(storage)
    assert config is None

    # Test create_async with custom config
    custom_config = ic.RepositoryConfig.default()
    custom_config.inline_chunk_threshold_bytes = 1024

    repo = await ic.Repository.create_async(
        storage=storage,
        config=custom_config,
    )

    # Test exists_async with existing repo
    assert await ic.Repository.exists_async(storage)

    # Test fetch_config_async with existing repo
    fetched_config = await ic.Repository.fetch_config_async(storage)
    assert fetched_config is not None
    assert fetched_config.inline_chunk_threshold_bytes == 1024

    # Test save_config_async - modify and save config
    await repo.save_config_async()

    # Verify config was saved
    saved_config = await ic.Repository.fetch_config_async(storage)
    assert saved_config is not None
    assert saved_config.inline_chunk_threshold_bytes == 1024

    # Test open_async
    opened_repo = await ic.Repository.open_async(storage=storage)
    assert opened_repo.config.inline_chunk_threshold_bytes == 1024

    # Test reopen_async with new config
    new_config = ic.RepositoryConfig.default()
    new_config.inline_chunk_threshold_bytes = 4096

    reopened_repo = await repo.reopen_async(config=new_config)
    assert reopened_repo.config.inline_chunk_threshold_bytes == 4096

    # Test reopen (sync) with new config
    new_config_sync = ic.RepositoryConfig.default()
    new_config_sync.inline_chunk_threshold_bytes = 2048

    reopened_repo_sync = repo.reopen(config=new_config_sync)
    assert reopened_repo_sync.config.inline_chunk_threshold_bytes == 2048

    # Test open_or_create_async with new storage (should create)
    new_storage = ic.in_memory_storage()
    assert not await ic.Repository.exists_async(new_storage)


async def test_rewrite_manifests_async() -> None:
    """Test Repository.rewrite_manifests_async method."""
    repo = await ic.Repository.create_async(
        storage=ic.in_memory_storage(),
        config=ic.RepositoryConfig(inline_chunk_threshold_bytes=0),
    )

    # Create some data to generate manifests
    session = await repo.writable_session_async("main")
    group = zarr.group(store=session.store, overwrite=True)
    array = group.create_array(
        "test_array",
        shape=(100, 50),
        chunks=(10, 10),
        dtype="i4",
        compressors=None,
    )

    # Write data to create chunks and manifests
    array[:50, :25] = 42
    _first_commit = await session.commit_async("initial data")

    # Add more data
    session = await repo.writable_session_async("main")
    group = zarr.open_group(store=session.store)
    array = group["test_array"]
    array[50:, 25:] = 99
    second_commit = await session.commit_async("more data")

    # Get initial ancestry to verify commits exist
    initial_ancestry = [snap async for snap in repo.async_ancestry(branch="main")]
    assert len(initial_ancestry) >= 3  # initial + first_commit + second_commit

    # Test rewrite_manifests_async
    rewrite_commit = await repo.rewrite_manifests_async(
        "rewritten manifests",
        branch="main",
    )

    # Verify the rewrite created a new commit
    assert rewrite_commit != second_commit

    # Verify ancestry after rewrite
    new_ancestry = [snap async for snap in repo.async_ancestry(branch="main")]
    assert len(new_ancestry) == len(initial_ancestry) + 1
    assert new_ancestry[0].message == "rewritten manifests"

    # Verify data is still accessible after manifest rewrite
    session = await repo.readonly_session_async("main")
    group = zarr.open_group(store=session.store, mode="r")
    array = group["test_array"]

    # Check that data is preserved
    assert array[0, 0] == 42  # from first commit
    assert array[99, 49] == 99  # from second commit



import pickle
from collections.abc import Callable, Mapping
from datetime import datetime
from typing import cast

from icechunk._icechunk_python import (
    AzureCredentials,
    AzureStaticCredentials,
    Credentials,
    GcsBearerCredential,
    GcsCredentials,
    GcsStaticCredentials,
    S3Credentials,
    S3StaticCredentials,
)

AnyS3Credential = (
    S3Credentials.Static
    | S3Credentials.Anonymous
    | S3Credentials.FromEnv
    | S3Credentials.Refreshable
)

AnyGcsStaticCredential = (
    GcsStaticCredentials.ServiceAccount
    | GcsStaticCredentials.ServiceAccountKey
    | GcsStaticCredentials.ApplicationCredentials
    | GcsStaticCredentials.BearerToken
)

AnyGcsCredential = (
    GcsCredentials.FromEnv | GcsCredentials.Static | GcsCredentials.Refreshable
)

AnyAzureStaticCredential = (
    AzureStaticCredentials.AccessKey
    | AzureStaticCredentials.SasToken
    | AzureStaticCredentials.BearerToken
)

AnyAzureCredential = AzureCredentials.FromEnv | AzureCredentials.Static


AnyCredential = Credentials.S3 | Credentials.Gcs | Credentials.Azure


def s3_refreshable_credentials(
    get_credentials: Callable[[], S3StaticCredentials],
    scatter_initial_credentials: bool = False,
) -> S3Credentials.Refreshable:
    """Create refreshable credentials for S3 and S3 compatible object stores.

    Parameters
    ----------
    get_credentials: Callable[[], S3StaticCredentials]
        Use this function to get and refresh the credentials. The function must be pickable.
    scatter_initial_credentials: bool, optional
        Immediately call and store the value returned by get_credentials. This is useful if the
        repo or session will be pickled to generate many copies. Passing scatter_initial_credentials=True will
        ensure all those copies don't need to call get_credentials immediately. After the initial
        set of credentials has expired, the cached value is no longer used. Notice that credentials
        obtained are stored, and they can be sent over the network if you pickle the session/repo.
    """
    current = get_credentials() if scatter_initial_credentials else None
    return S3Credentials.Refreshable(pickle.dumps(get_credentials), current)


def s3_static_credentials(
    *,
    access_key_id: str,
    secret_access_key: str,
    session_token: str | None = None,
    expires_after: datetime | None = None,
) -> S3Credentials.Static:
    """Create static credentials for S3 and S3 compatible object stores.

    Parameters
    ----------
    access_key_id: str | None
        S3 credential access key
    secret_access_key: str | None
        S3 credential secret access key
    session_token: str | None
        Optional S3 credential session token
    expires_after: datetime | None
        Optional expiration for the object store credentials
    """
    return S3Credentials.Static(
        S3StaticCredentials(
            access_key_id=access_key_id,
            secret_access_key=secret_access_key,
            session_token=session_token,
            expires_after=expires_after,
        )
    )


def s3_anonymous_credentials() -> S3Credentials.Anonymous:
    """Create no-signature credentials for S3 and S3 compatible object stores."""
    return S3Credentials.Anonymous()


def s3_from_env_credentials() -> S3Credentials.FromEnv:
    """Instruct S3 and S3 compatible object stores to gather credentials from the operative system environment."""
    return S3Credentials.FromEnv()


def s3_credentials(
    *,
    access_key_id: str | None = None,
    secret_access_key: str | None = None,
    session_token: str | None = None,
    expires_after: datetime | None = None,
    anonymous: bool | None = None,
    from_env: bool | None = None,
    get_credentials: Callable[[], S3StaticCredentials] | None = None,
    scatter_initial_credentials: bool = False,
) -> AnyS3Credential:
    """Create credentials for S3 and S3 compatible object stores.

    If all arguments are None, credentials are fetched from the environment.

    Parameters
    ----------
    access_key_id: str | None
        S3 credential access key
    secret_access_key: str | None
        S3 credential secret access key
    session_token: str | None
        Optional S3 credential session token
    expires_after: datetime | None
        Optional expiration for the object store credentials
    anonymous: bool | None
        If set to True requests to the object store will not be signed
    from_env: bool | None
        Fetch credentials from the operative system environment
    get_credentials: Callable[[], S3StaticCredentials] | None
        Use this function to get and refresh object store credentials
    scatter_initial_credentials: bool, optional
        Immediately call and store the value returned by get_credentials. This is useful if the
        repo or session will be pickled to generate many copies. Passing scatter_initial_credentials=True will
        ensure all those copies don't need to call get_credentials immediately. After the initial
        set of credentials has expired, the cached value is no longer used. Notice that credentials
        obtained are stored, and they can be sent over the network if you pickle the session/repo.
    """
    if (
        (from_env is None or from_env)
        and access_key_id is None
        and secret_access_key is None
        and session_token is None
        and expires_after is None
        and not anonymous
        and get_credentials is None
    ):
        return s3_from_env_credentials()

    if (
        anonymous
        and access_key_id is None
        and secret_access_key is None
        and session_token is None
        and expires_after is None
        and not from_env
        and get_credentials is None
    ):
        return s3_anonymous_credentials()

    if (
        get_credentials is not None
        and access_key_id is None
        and secret_access_key is None
        and session_token is None
        and expires_after is None
        and not from_env
        and not anonymous
    ):
        return s3_refreshable_credentials(
            get_credentials, scatter_initial_credentials=scatter_initial_credentials
        )

    if (
        access_key_id
        and secret_access_key
        and not from_env
        and not anonymous
        and get_credentials is None
    ):
        return s3_static_credentials(
            access_key_id=access_key_id,
            secret_access_key=secret_access_key,
            session_token=session_token,
            expires_after=expires_after,
        )

    raise ValueError("Conflicting arguments to s3_credentials function")


def gcs_static_credentials(
    *,
    service_account_file: str | None = None,
    service_account_key: str | None = None,
    application_credentials: str | None = None,
    bearer_token: str | None = None,
) -> AnyGcsStaticCredential:
    """Create static credentials Google Cloud Storage object store."""
    if service_account_file is not None:
        return GcsStaticCredentials.ServiceAccount(service_account_file)
    if service_account_key is not None:
        return GcsStaticCredentials.ServiceAccountKey(service_account_key)
    if application_credentials is not None:
        return GcsStaticCredentials.ApplicationCredentials(application_credentials)
    if bearer_token is not None:
        return GcsStaticCredentials.BearerToken(bearer_token)
    raise ValueError("Conflicting arguments to gcs_static_credentials function")


def gcs_refreshable_credentials(
    get_credentials: Callable[[], GcsBearerCredential],
    scatter_initial_credentials: bool = False,
) -> GcsCredentials.Refreshable:
    """Create refreshable credentials for Google Cloud Storage object store.

    Parameters
    ----------
    get_credentials: Callable[[], S3StaticCredentials]
        Use this function to get and refresh the credentials. The function must be pickable.
    scatter_initial_credentials: bool, optional
        Immediately call and store the value returned by get_credentials. This is useful if the
        repo or session will be pickled to generate many copies. Passing scatter_initial_credentials=True will
        ensure all those copies don't need to call get_credentials immediately. After the initial
        set of credentials has expired, the cached value is no longer used. Notice that credentials
        obtained are stored, and they can be sent over the network if you pickle the session/repo.
    """

    current = get_credentials() if scatter_initial_credentials else None
    return GcsCredentials.Refreshable(pickle.dumps(get_credentials), current)


def gcs_from_env_credentials() -> GcsCredentials.FromEnv:
    """Instruct Google Cloud Storage object store to fetch credentials from the operative system environment."""
    return GcsCredentials.FromEnv()


def gcs_credentials(
    *,
    service_account_file: str | None = None,
    service_account_key: str | None = None,
    application_credentials: str | None = None,
    bearer_token: str | None = None,
    from_env: bool | None = None,
    get_credentials: Callable[[], GcsBearerCredential] | None = None,
    scatter_initial_credentials: bool = False,
) -> AnyGcsCredential:
    """Create credentials Google Cloud Storage object store.

    If all arguments are None, credentials are fetched from the operative system environment.
    """
    if (from_env is None or from_env) and (
        service_account_file is None
        and service_account_key is None
        and application_credentials is None
        and bearer_token is None
    ):
        return gcs_from_env_credentials()

    if (
        service_account_file is not None
        or service_account_key is not None
        or application_credentials is not None
        or bearer_token is not None
    ) and (from_env is None or not from_env):
        return GcsCredentials.Static(
            gcs_static_credentials(
                service_account_file=service_account_file,
                service_account_key=service_account_key,
                application_credentials=application_credentials,
                bearer_token=bearer_token,
            )
        )

    if get_credentials is not None:
        return gcs_refreshable_credentials(
            get_credentials, scatter_initial_credentials=scatter_initial_credentials
        )

    raise ValueError("Conflicting arguments to gcs_credentials function")


def azure_static_credentials(
    *,
    access_key: str | None = None,
    sas_token: str | None = None,
    bearer_token: str | None = None,
) -> AnyAzureStaticCredential:
    """Create static credentials Azure Blob Storage object store."""
    if [access_key, sas_token, bearer_token].count(None) != 2:
        raise ValueError("Conflicting arguments to azure_static_credentials function")
    if access_key is not None:
        return AzureStaticCredentials.AccessKey(access_key)
    if sas_token is not None:
        return AzureStaticCredentials.SasToken(sas_token)
    if bearer_token is not None:
        return AzureStaticCredentials.BearerToken(bearer_token)
    raise ValueError(
        "No valid static credential provided for Azure Blob Storage object store"
    )


def azure_from_env_credentials() -> AzureCredentials.FromEnv:
    """Instruct Azure Blob Storage object store to fetch credentials from the operative system environment."""
    return AzureCredentials.FromEnv()


def azure_credentials(
    *,
    access_key: str | None = None,
    sas_token: str | None = None,
    bearer_token: str | None = None,
    from_env: bool | None = None,
) -> AnyAzureCredential:
    """Create credentials Azure Blob Storage object store.

    If all arguments are None, credentials are fetched from the operative system environment.
    """
    if (from_env is None or from_env) and (
        access_key is None and sas_token is None and bearer_token is None
    ):
        return azure_from_env_credentials()

    if (access_key is not None or sas_token is not None or bearer_token is not None) and (
        from_env is None or not from_env
    ):
        return AzureCredentials.Static(
            azure_static_credentials(
                access_key=access_key,
                sas_token=sas_token,
                bearer_token=bearer_token,
            )
        )

    raise ValueError("Conflicting arguments to azure_credentials function")


def containers_credentials(
    m: Mapping[str, AnyS3Credential | AnyGcsCredential | AnyAzureCredential | None],
) -> dict[str, AnyCredential | None]:
    """Build a map of credentials for virtual chunk containers.

    Parameters
    ----------
    m: Mapping[str, AnyS3Credential | AnyGcsCredential | AnyAzureCredential ]
        A mapping from container url prefixes to credentials.

    Examples
    --------
    ```python
    import icechunk as ic

    config = ic.RepositoryConfig.default()
    config.inline_chunk_threshold_bytes = 512

    virtual_store_config = ic.s3_store(
        region="us-east-1",
        endpoint_url="http://localhost:9000",
        allow_http=True,
        s3_compatible=True,
        force_path_style=True,
    )
    container = ic.VirtualChunkContainer("s3://somebucket", virtual_store_config)
    config.set_virtual_chunk_container(container)
    credentials = ic.containers_credentials(
        {"s3://somebucket": ic.s3_credentials(access_key_id="ACCESS_KEY", secret_access_key="SECRET"}
    )

    repo = ic.Repository.create(
        storage=ic.local_filesystem_storage(store_path),
        config=config,
        authorize_virtual_chunk_access=credentials,
    )
    ```

    """
    res: dict[str, AnyCredential | None] = {}
    for name, cred in m.items():
        if cred is None:
            res[name] = None
        elif isinstance(cred, AnyS3Credential):
            res[name] = Credentials.S3(cred)
        elif (
            isinstance(cred, GcsCredentials.FromEnv)
            or isinstance(cred, GcsCredentials.Static)
            or isinstance(cred, GcsCredentials.Refreshable)
        ):
            res[name] = Credentials.Gcs(cast(GcsCredentials, cred))
        elif isinstance(cred, AzureCredentials.FromEnv) or isinstance(
            cred, AzureCredentials.Static
        ):
            res[name] = Credentials.Azure(cast(AzureCredentials, cred))
        else:
            raise ValueError(f"Unknown credential type {type(cred)}")
    return res



from pathlib import Path
from typing import cast

import numpy as np
import pytest

import icechunk
import zarr


@pytest.fixture
def repo(tmpdir: Path) -> icechunk.Repository:
    repo = icechunk.Repository.create(
        storage=icechunk.local_filesystem_storage(str(tmpdir))
    )

    session = repo.writable_session("main")
    store = session.store
    root = zarr.group(store=store)
    root.create_group("foo/bar")
    root.create_array("foo/bar/some-array", shape=(10, 10), chunks=(1, 1), dtype="i4")
    session.commit("commit 1")

    return repo


def test_detect_conflicts(repo: icechunk.Repository) -> None:
    session_a = repo.writable_session("main")
    session_b = repo.writable_session("main")
    store_a = session_a.store
    store_b = session_b.store

    root_a = zarr.group(store=store_a)
    array_a = cast(zarr.Array, root_a["foo/bar/some-array"])
    array_a[:, :] = 1
    array_a.attrs["repo"] = 1
    session_a.commit("update array")

    root_b = zarr.group(store=store_b)
    array_b = cast(zarr.Array, root_b["foo/bar/some-array"])
    array_b[:, :] = 2
    array_b.attrs["repo"] = 2

    with pytest.raises(icechunk.ConflictError):
        session_b.commit("update array")

    with pytest.raises(icechunk.RebaseFailedError):
        try:
            session_b.rebase(icechunk.ConflictDetector())
        except icechunk.RebaseFailedError as e:
            assert len(e.conflicts) == 3
            assert e.conflicts[0].path == "/foo/bar/some-array"
            assert e.conflicts[0].path == "/foo/bar/some-array"
            assert (
                e.conflicts[0].conflict_type
                == icechunk.ConflictType.ZarrMetadataDoubleUpdate
            )
            assert e.conflicts[1].path == "/foo/bar/some-array"
            assert (
                e.conflicts[1].conflict_type
                == icechunk.ConflictType.ChunksUpdatedInUpdatedArray
            )
            assert e.conflicts[2].path == "/foo/bar/some-array"
            assert e.conflicts[2].conflict_type == icechunk.ConflictType.ChunkDoubleUpdate
            assert e.conflicts[2].conflicted_chunks
            assert len(e.conflicts[2].conflicted_chunks) == 100

            raise e


def test_rebase_no_conflicts(repo: icechunk.Repository) -> None:
    session_a = repo.writable_session("main")
    session_b = repo.writable_session("main")
    store_a = session_a.store
    store_b = session_b.store

    root_a = zarr.group(store=store_a)
    array_a = cast(zarr.Array, root_a["foo/bar/some-array"])
    array_a[:] = 1
    session_a.commit("update array")

    root_b = zarr.group(store=store_b)
    array_b = cast(zarr.Array, root_b["foo/bar/some-array"])
    array_b.attrs["repo"] = 2

    session_b.commit("update array", rebase_with=icechunk.ConflictDetector())

    session_c = repo.readonly_session(branch="main")
    store_c = session_c.store
    root_c = zarr.open_group(store=store_c, mode="r")
    array_c = cast(zarr.Array, root_c["foo/bar/some-array"])
    np.testing.assert_array_equal(array_c[:], 1)
    assert array_c.attrs["repo"] == 2


def test_rebase_fails_on_user_atts_double_edit(repo: icechunk.Repository) -> None:
    session_a = repo.writable_session("main")
    session_b = repo.writable_session("main")
    store_a = session_a.store
    store_b = session_b.store

    root_a = zarr.group(store=store_a)
    array_a = cast(zarr.Array, root_a["foo/bar/some-array"])
    array_a.attrs["repo"] = 1
    session_a.commit("update array")

    root_b = zarr.group(store=store_b)
    array_b = cast(zarr.Array, root_b["foo/bar/some-array"])
    array_b.attrs["repo"] = 2

    with pytest.raises(icechunk.ConflictError):
        session_b.commit("update array")

    # Make sure it fails if the resolver is not set
    with pytest.raises(icechunk.RebaseFailedError):
        session_b.commit("update array", rebase_with=icechunk.BasicConflictSolver())


@pytest.mark.parametrize(
    "on_chunk_conflict",
    [icechunk.VersionSelection.UseOurs, icechunk.VersionSelection.UseTheirs],
)
def test_rebase_chunks_with_ours(
    repo: icechunk.Repository, on_chunk_conflict: icechunk.VersionSelection
) -> None:
    session_a = repo.writable_session("main")
    session_b = repo.writable_session("main")
    store_a = session_a.store
    store_b = session_b.store

    root_a = zarr.group(store=store_a)
    array_a = cast(zarr.Array, root_a["foo/bar/some-array"])
    array_a[:] = 1
    session_a.commit("update array")

    root_b = zarr.group(store=store_b)
    array_b = cast(zarr.Array, root_b["foo/bar/some-array"])
    array_b[:, 0] = 2

    with pytest.raises(icechunk.ConflictError):
        session_b.commit("update first column of array")

    # Make sure it fails if the resolver is not set
    with pytest.raises(icechunk.RebaseFailedError):
        try:
            session_b.commit(
                "update first column of array",
                rebase_with=icechunk.BasicConflictSolver(
                    on_chunk_conflict=icechunk.VersionSelection.Fail
                ),
            )
        except icechunk.RebaseFailedError as e:
            assert e.conflicts[0].path == "/foo/bar/some-array"
            assert e.conflicts[0].conflict_type == icechunk.ConflictType.ChunkDoubleUpdate
            assert len(e.conflicts) == 1

            np.testing.assert_array_equal(
                np.array(e.conflicts[0].conflicted_chunks),
                np.array(
                    [
                        [0, 0],
                        [1, 0],
                        [2, 0],
                        [3, 0],
                        [4, 0],
                        [5, 0],
                        [6, 0],
                        [7, 0],
                        [8, 0],
                        [9, 0],
                    ]
                ),
            )

            raise e

    solver = icechunk.BasicConflictSolver(
        on_chunk_conflict=on_chunk_conflict,
    )

    session_b.commit("after conflict", rebase_with=solver)

    session_c = repo.readonly_session(branch="main")
    store_c = session_c.store
    root_c = zarr.open_group(store=store_c, mode="r")
    array_c = cast(zarr.Array, root_c["foo/bar/some-array"])
    assert (
        array_c[0, 0] == 2
        if on_chunk_conflict == icechunk.VersionSelection.UseOurs
        else 1
    )
    assert array_c[0, 1] == 1


async def test_rebase_async() -> None:
    """Test async rebase functionality with conflict detection and resolution."""
    repo = await icechunk.Repository.create_async(
        storage=icechunk.in_memory_storage(),
    )

    # Set up initial state
    session_a = await repo.writable_session_async("main")
    root_a = zarr.group(store=session_a.store, overwrite=True)
    group_a = root_a.create_group("foo/bar")
    array_a = group_a.create_array(
        "some-array", shape=(100, 100), chunks=(10, 10), dtype="i4"
    )
    array_a[:, :] = 1
    array_a.attrs["repo"] = 1
    await session_a.commit_async("create array")

    # Create conflicting sessions
    session_b = await repo.writable_session_async("main")
    root_b = zarr.open_group(store=session_b.store)
    array_b = cast(zarr.Array, root_b["foo/bar/some-array"])

    session_c = await repo.writable_session_async("main")
    root_c = zarr.open_group(store=session_c.store)
    array_c = cast(zarr.Array, root_c["foo/bar/some-array"])

    # Make changes in session_c and commit first
    array_c[:, :] = 3
    await session_c.commit_async("update array from c")

    # Make conflicting changes in session_b
    array_b[:, :] = 2

    # This should fail due to conflicts
    with pytest.raises(icechunk.ConflictError):
        await session_b.commit_async("update array")

    # Test rebase_async with ConflictDetector (should fail)
    with pytest.raises(icechunk.RebaseFailedError):
        try:
            await session_b.rebase_async(icechunk.ConflictDetector())
        except icechunk.RebaseFailedError as e:
            assert len(e.conflicts) == 1
            assert e.conflicts[0].path == "/foo/bar/some-array"
            assert e.conflicts[0].conflict_type == icechunk.ConflictType.ChunkDoubleUpdate
            assert e.conflicts[0].conflicted_chunks
            assert len(e.conflicts[0].conflicted_chunks) == 100
            raise e

    # Test successful rebase_async with BasicConflictSolver
    solver = icechunk.BasicConflictSolver(
        on_chunk_conflict=icechunk.VersionSelection.UseOurs,
    )

    await session_b.rebase_async(solver)
    await session_b.commit_async("after conflict resolution")

    # Verify the result
    session_final = await repo.readonly_session_async(branch="main")
    store_final = session_final.store
    root_final = zarr.open_group(store=store_final, mode="r")
    array_final = cast(zarr.Array, root_final["foo/bar/some-array"])

    # Should have session_b's values due to UseOurs selection
    assert array_final[0, 0] == 2
    assert array_final.attrs["repo"] == 1



import operator
import time
from contextlib import contextmanager
from dataclasses import dataclass, field
from itertools import product, repeat

import numpy as np
import pandas as pd
import toolz as tlz


def stats(data):
    q = np.quantile(data, q=[0.25, 0.5, 0.75])
    _, trend = np.polyfit(np.arange(data.size), data, deg=1)
    return {
        "rounds": data.size,
        "max": data.max(),
        "trend": trend,
        "mean": data.mean(),
        "iqr": q[2] - q[0],
        "min": data.min(),
        "median": q[1],
        "q1": q[0],
        "q3": q[2],
        "stddev": np.std(data),
    }


def slices_from_chunks(shape: tuple[int, ...], chunks: tuple[int, ...]):
    """slightly modified from dask.array.core.slices_from_chunks to be lazy"""

    extras = ((s % c,) if s % c > 0 else () for s, c in zip(shape, chunks, strict=True))
    # need this twice
    chunks = tuple(
        tuple(tlz.concatv(repeat(c, s // c), e))
        for s, c, e in zip(shape, chunks, extras, strict=True)
    )
    cumdims = (tlz.accumulate(operator.add, bds[:-1], 0) for bds in chunks)
    slices = (
        (slice(s, s + dim) for s, dim in zip(starts, shapes, strict=True))
        for starts, shapes in zip(cumdims, chunks, strict=True)
    )
    return product(*slices)


def normalize_chunks(
    *, shape: tuple[int, ...], chunks: tuple[int, ...]
) -> tuple[int, ...]:
    assert len(shape) == len(chunks)
    chunks = tuple(s if c == -1 else c for s, c in zip(shape, chunks, strict=True))
    return chunks


def get_task_chunk_shape(
    *, task_nchunks: int, shape: tuple[int, ...], chunks: tuple[int, ...]
) -> tuple[int, ...]:
    left = task_nchunks
    task_chunk_shape = []
    for s, c in zip(shape, chunks, strict=True):
        if c == s or left is None:
            task_chunk_shape.append(c)
        else:
            q, r = divmod(s, c)
            if q > left:
                task_chunk_shape.append(left * c)
            else:
                task_chunk_shape.append(q * c)
                left /= q
    print(f"{task_chunk_shape=!r}")
    return task_chunk_shape


@dataclass
class Timer:
    diagnostics: list = field(default_factory=list)

    @contextmanager
    def time(self, **kwargs):
        tic = time.perf_counter()
        yield
        toc = time.perf_counter()
        kwargs["runtime"] = toc - tic
        self.diagnostics.append(kwargs)

    def as_dict(self) -> dict:
        out = {}
        for item in self.diagnostics:
            out[item["op"]] = item["runtime"]
        return out

    def dataframe(self):
        return pd.DataFrame(self.diagnostics)



import functools
import json
import pickle
from typing import Any

import hypothesis.extra.numpy as npst
import hypothesis.strategies as st
import numpy as np
import pytest
from hypothesis import assume, note
from hypothesis.stateful import (
    Settings,
    invariant,
    precondition,
    rule,
    run_state_machine_as_test,
)

import icechunk as ic
import zarr
from icechunk import Repository, Storage, in_memory_storage
from icechunk.testing import strategies as icst
from zarr.core.buffer import default_buffer_prototype
from zarr.testing.stateful import ZarrHierarchyStateMachine
from zarr.testing.strategies import (
    basic_indices,
    node_names,
    np_array_and_chunks,
    orthogonal_indices,
)

PROTOTYPE = default_buffer_prototype()

# pytestmark = [
#     pytest.mark.filterwarnings(
#         "ignore::zarr.core.dtype.common.UnstableSpecificationWarning"
#     ),
# ]


def with_frequency(frequency):
    """
    Decorator to control how frequently a rule runs in Hypothesis stateful tests.

    Args:
        frequency: Float between 0 and 1, where 1.0 means always run,
                  0.1 means run ~10% of the time, etc.

    Usage:
        @rule()
        @with_frequency(0.1)  # Run ~10% of the time
        def rare_operation(self):
            pass
    """

    def decorator(func):
        # Create a counter attribute name specific to this function
        counter_attr = f"__{func.__name__}_counter"

        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            return func(self, *args, **kwargs)

        # Add precondition that checks frequency
        @precondition
        def frequency_check(self):
            # Initialize counter if it doesn't exist
            if not hasattr(self, counter_attr):
                setattr(self, counter_attr, 0)

            # Increment counter
            current_count = getattr(self, counter_attr) + 1
            setattr(self, counter_attr, current_count)

            # Check if we should run based on frequency
            # This gives roughly the right frequency over many calls
            return (current_count * frequency) % 1.0 >= (1.0 - frequency)

        # Apply the precondition to the wrapped function
        return frequency_check(wrapper)

    return decorator


@st.composite
def chunk_paths(
    draw: st.DrawFn, ndim: int, numblocks: tuple[int, ...], subset: bool = True
) -> str:
    blockidx = draw(
        st.tuples(*tuple(st.integers(min_value=0, max_value=b - 1) for b in numblocks))
    )
    subset_slicer = (
        slice(draw(st.integers(min_value=0, max_value=ndim))) if subset else slice(None)
    )
    return "/".join(map(str, blockidx[subset_slicer]))


# TODO: more before/after commit invariants?
# TODO: add "/" to self.all_groups, deleting "/" seems to be problematic
class ModifiedZarrHierarchyStateMachine(ZarrHierarchyStateMachine):
    def __init__(self, storage: Storage) -> None:
        self.storage = storage
        self.repo = Repository.create(self.storage)
        store = self.repo.writable_session("main").store
        super().__init__(store)

    @precondition(
        lambda self: not self.store.session.has_uncommitted_changes
        and bool(self.all_arrays)
    )
    @rule(data=st.data())
    def reopen_with_config(self, data):
        array_paths = data.draw(
            st.lists(st.sampled_from(sorted(self.all_arrays)), max_size=3, unique=True)
        )
        arrays = tuple(zarr.open_array(self.model, path=path) for path in array_paths)
        sconfig = data.draw(icst.splitting_configs(arrays=arrays))
        config = ic.RepositoryConfig(
            inline_chunk_threshold_bytes=0, manifest=ic.ManifestConfig(splitting=sconfig)
        )
        note(f"reopening with splitting config {sconfig=!r}")
        self.repo = Repository.open(self.storage, config=config)
        if data.draw(st.booleans()):
            self.repo.save_config()
        self.store = self.repo.writable_session("main").store

    @precondition(lambda self: not self.store.session.has_uncommitted_changes)
    @rule(data=st.data())
    def rewrite_manifests(self, data):
        config_dict = {
            ic.ManifestSplitCondition.AnyArray(): {
                ic.ManifestSplitDimCondition.Any(): data.draw(
                    st.integers(min_value=1, max_value=10)
                )
            }
        }
        sconfig = ic.ManifestSplittingConfig.from_dict(config_dict)

        config = ic.RepositoryConfig(
            inline_chunk_threshold_bytes=0, manifest=ic.ManifestConfig(splitting=sconfig)
        )
        note(f"rewriting manifests with config {sconfig=!r}")
        self.repo = Repository.open(self.storage, config=config)
        self.repo.rewrite_manifests(
            f"rewriting manifests with {sconfig!s}", branch="main"
        )
        if data.draw(st.booleans()):
            self.repo.save_config()
        self.store = self.repo.writable_session("main").store

    @precondition(lambda self: self.store.session.has_uncommitted_changes)
    @rule(data=st.data())
    def commit_with_check(self, data) -> None:
        note("committing and checking list_prefix")

        lsbefore = sorted(self._sync_iter(self.store.list_prefix("")))
        path = data.draw(st.sampled_from(lsbefore))
        get_before = self._sync(self.store.get(path, prototype=PROTOTYPE))
        assert get_before

        self.store.session.commit("foo")

        self.store = self.repo.writable_session("main").store

        lsafter = sorted(self._sync_iter(self.store.list_prefix("")))
        get_after = self._sync(self.store.get(path, prototype=PROTOTYPE))
        assert get_after

        if lsbefore != lsafter:
            lsexpect = sorted(self._sync_iter(self.model.list_prefix("")))
            raise ValueError(
                f"listing changed before ({len(lsbefore)} items) and after ({len(lsafter)} items) committing."
                f" \n\n Before : {lsbefore!r} \n\n After: {lsafter!r}, \n\n Expected: {lsexpect!r}"
            )

        # if it's metadata, we need to compare the data parsed, not raw (because of map ordering)
        if path.endswith(".json"):
            get_after = json.loads(get_after.to_bytes())
            get_before = json.loads(get_before.to_bytes())
        else:
            get_after = get_after.to_bytes()
            get_before = get_before.to_bytes()

        if get_before != get_after:
            get_expect = self._sync(self.model.get(path, prototype=PROTOTYPE))
            assert get_expect
            raise ValueError(
                f"Value changed before and after commit for path {path}"
                f" \n\n Before : {get_before!r} \n\n "
                f"After: {get_after!r}, \n\n "
                f"Expected: {get_expect.to_bytes()!r}"
            )

    @rule(
        data=st.data(),
        name=node_names,
        array_and_chunks=np_array_and_chunks(),
    )
    def add_array(
        self,
        data: st.DataObject,
        name: str,
        array_and_chunks: tuple[np.ndarray[Any, Any], tuple[int, ...]],
    ) -> None:
        array, _ = array_and_chunks
        # TODO: support size-0 arrays GH392
        assume(array.size > 0)
        super().add_array(data, name, array_and_chunks)

    @precondition(lambda self: bool(self.all_groups))
    @rule(data=st.data())
    def check_list_dir(self, data: st.DataObject) -> None:
        path = self.draw_directory(data)
        note(f"list_dir for {path=!r}")
        model_ls = sorted(self._sync_iter(self.model.list_dir(path)))
        store_ls = sorted(self._sync_iter(self.store.list_dir(path)))
        if model_ls != store_ls and set(model_ls).symmetric_difference(set(store_ls)) != {
            "c"
        }:
            # Consider .list_dir("path/to/array") for an array with a single chunk.
            # The MemoryStore model will return `"c", "zarr.json"` only if the chunk exists
            # If that chunk was deleted, then `"c"` is not returned.
            # LocalStore will not have this behaviour :/
            # In Icechunk, we always return the `c` so ignore this inconsistency.
            assert model_ls == store_ls, (model_ls, store_ls)

    #####  TODO: port everything below to zarr
    @precondition(lambda self: bool(self.all_arrays))
    @rule(data=st.data())
    def check_array(self, data: st.DataObject) -> None:
        path = data.draw(st.sampled_from(sorted(self.all_arrays)))
        actual = zarr.open_array(self.store, path=path)[:]
        expected = zarr.open_array(self.model, path=path)[:]
        np.testing.assert_equal(actual, expected)

    @precondition(lambda self: bool(self.all_arrays))
    @rule(data=st.data())
    def overwrite_array_orthogonal_indexing(self, data: st.DataObject) -> None:
        array = data.draw(st.sampled_from(sorted(self.all_arrays)))
        model_array = zarr.open_array(path=array, store=self.model)
        store_array = zarr.open_array(path=array, store=self.store)
        indexer, _ = data.draw(orthogonal_indices(shape=model_array.shape))
        note(f"overwriting array orthogonal {indexer=}")
        new_data = data.draw(
            npst.arrays(shape=model_array.oindex[indexer].shape, dtype=model_array.dtype)
        )
        model_array.oindex[indexer] = new_data
        store_array.oindex[indexer] = new_data

    #####  TODO: delete after next Zarr release (Jun 18, 2025)
    @rule()
    @with_frequency(0.25)
    def clear(self) -> None:
        note("clearing")
        import zarr

        self._sync(self.store.clear())
        self._sync(self.model.clear())

        assert self._sync(self.store.is_empty("/"))
        assert self._sync(self.model.is_empty("/"))

        self.all_groups.clear()
        self.all_arrays.clear()

        zarr.group(store=self.store)
        zarr.group(store=self.model)

        assert not self._sync(self.store.is_empty("/"))
        # TODO: MemoryStore is broken?
        # assert not self._sync(self.model.is_empty("/"))

    def draw_directory(self, data) -> str:
        group_st = (
            st.sampled_from(sorted(self.all_groups)) if self.all_groups else st.nothing()
        )
        array_st = (
            st.sampled_from(sorted(self.all_arrays)) if self.all_arrays else st.nothing()
        )
        array_or_group = data.draw(st.one_of(group_st, array_st))
        if data.draw(st.booleans()) and array_or_group in self.all_arrays:
            arr = zarr.open_array(path=array_or_group, store=self.model)
            path = data.draw(
                st.one_of(
                    st.sampled_from([array_or_group]),
                    chunk_paths(ndim=arr.ndim, numblocks=arr.cdata_shape).map(
                        lambda x: f"{array_or_group}/c/"
                    ),
                )
            )
        else:
            path = array_or_group
        return path

    @precondition(lambda self: bool(self.all_arrays))
    @rule(data=st.data())
    def delete_chunk(self, data) -> None:
        array = data.draw(st.sampled_from(sorted(self.all_arrays)))
        arr = zarr.open_array(path=array, store=self.model)
        chunk_path = data.draw(
            chunk_paths(ndim=arr.ndim, numblocks=arr.cdata_shape, subset=False)
        )
        path = f"{array}/c/{chunk_path}"
        note(f"deleting chunk {path=!r}")
        self._sync(self.model.delete(path))
        self._sync(self.store.delete(path))

    @precondition(lambda self: bool(self.all_arrays))
    @rule(data=st.data())
    def overwrite_array_basic_indexing(self, data) -> None:
        array = data.draw(st.sampled_from(sorted(self.all_arrays)))
        model_array = zarr.open_array(path=array, store=self.model)
        store_array = zarr.open_array(path=array, store=self.store)
        slicer = data.draw(basic_indices(shape=model_array.shape))
        note(f"overwriting array basic {slicer=}")
        new_data = data.draw(
            npst.arrays(shape=model_array[slicer].shape, dtype=model_array.dtype)
        )
        model_array[slicer] = new_data
        store_array[slicer] = new_data

    @precondition(lambda self: bool(self.all_arrays))
    @rule(data=st.data())
    def resize_array(self, data) -> None:
        array = data.draw(st.sampled_from(sorted(self.all_arrays)))
        model_array = zarr.open_array(path=array, store=self.model)
        store_array = zarr.open_array(path=array, store=self.store)
        ndim = model_array.ndim
        new_shape = data.draw(npst.array_shapes(max_dims=ndim, min_dims=ndim, min_side=1))
        note(f"resizing array from {model_array.shape} to {new_shape}")
        model_array.resize(new_shape)
        store_array.resize(new_shape)

    @precondition(lambda self: bool(self.all_arrays) or bool(self.all_groups))
    @rule(data=st.data())
    def delete_dir(self, data) -> None:
        path = self.draw_directory(data)
        note(f"delete_dir with {path=!r}")
        self._sync(self.model.delete_dir(path))
        self._sync(self.store.delete_dir(path))

        matches = set()
        for node in self.all_groups | self.all_arrays:
            if node.startswith(path):
                matches.add(node)
        self.all_groups = self.all_groups - matches
        self.all_arrays = self.all_arrays - matches

    @rule()
    def pickle_objects(self) -> None:
        if not self.store.session.has_uncommitted_changes:
            session = self.store.session.fork()
            pickle.loads(pickle.dumps(session))

        pickle.loads(pickle.dumps(self.repo))

    @invariant()
    def check_list_prefix_from_root(self) -> None:
        model_list = self._sync_iter(self.model.list_prefix(""))
        store_list = self._sync_iter(self.store.list_prefix(""))
        note(f"Checking {len(model_list)} expected keys vs {len(store_list)} actual keys")
        assert sorted(model_list) == sorted(store_list), (
            sorted(model_list),
            sorted(store_list),
        )

        # check that our internal state matches that of the store and model
        assert all(
            f"{path}/zarr.json" in model_list
            for path in self.all_groups | self.all_arrays
        )
        assert all(
            f"{path}/zarr.json" in store_list
            for path in self.all_groups | self.all_arrays
        )


def test_zarr_hierarchy() -> None:
    def mk_test_instance_sync() -> ModifiedZarrHierarchyStateMachine:
        return ModifiedZarrHierarchyStateMachine(in_memory_storage())

    run_state_machine_as_test(
        mk_test_instance_sync, settings=Settings(report_multiple_bugs=False)
    )


def test_zarr_store() -> None:
    pytest.skip("icechunk is more strict about keys")
    # repo = Repository.create(in_memory_storage())
    # store = repo.writable_session("main").store

    # def mk_test_instance_sync() -> ZarrHierarchyStateMachine:
    #     return ZarrStoreStateMachine(store)

    # run_state_machine_as_test(
    #     mk_test_instance_sync, settings=Settings(report_multiple_bugs=False)
    # )



import pickle
import time
from pathlib import Path

import pytest

import zarr
from icechunk import (
    IcechunkError,
    Repository,
    RepositoryConfig,
    S3StaticCredentials,
    local_filesystem_storage,
    s3_storage,
)


def create_local_repo(path: str) -> Repository:
    repo = Repository.create(storage=local_filesystem_storage(path))
    repo.set_default_commit_metadata({"author": "test"})
    return repo


@pytest.fixture(scope="function")
def tmp_repo(tmpdir: Path) -> Repository:
    store_path = f"{tmpdir}"
    repo = create_local_repo(store_path)
    return repo


def test_pickle_repository(tmpdir: Path, tmp_repo: Repository) -> None:
    pickled = pickle.dumps(tmp_repo)
    roundtripped = pickle.loads(pickled)
    assert tmp_repo.list_branches() == roundtripped.list_branches()
    assert tmp_repo.default_commit_metadata() == roundtripped.default_commit_metadata()
    assert tmp_repo.default_commit_metadata() == {"author": "test"}

    storage = tmp_repo.storage
    assert (
        repr(storage)
        == f"ObjectStorage(backend=LocalFileSystemObjectStoreBackend(path={tmpdir}))"
    )


def test_pickle_read_only(tmp_repo: Repository) -> None:
    tmp_session = tmp_repo.writable_session(branch="main")
    tmp_store = tmp_session.store

    assert tmp_store._read_only is False
    assert tmp_session.read_only is False

    with pytest.raises(ValueError):
        pickle.loads(pickle.dumps(tmp_store))

    roundtripped = pickle.loads(pickle.dumps(tmp_session.fork().store))
    assert roundtripped._read_only is False

    roundtripped = pickle.loads(pickle.dumps(tmp_session.fork()))
    assert roundtripped.store._read_only is False

    with pytest.raises(ValueError):
        pickle.loads(pickle.dumps(tmp_session))
    # assert roundtripped.store._read_only is False


def get_credentials() -> S3StaticCredentials:
    return S3StaticCredentials("minio123", "minio123")


def test_pickle() -> None:
    # we test with refreshable credentials because that gave us problems in the past

    def mk_repo() -> tuple[str, Repository]:
        prefix = "test-repo__" + str(time.time())
        repo = Repository.create(
            storage=s3_storage(
                endpoint_url="http://localhost:9000",
                allow_http=True,
                force_path_style=True,
                region="us-east-1",
                bucket="testbucket",
                prefix=prefix,
                get_credentials=get_credentials,
            ),
            config=RepositoryConfig(inline_chunk_threshold_bytes=0),
        )
        return (prefix, repo)

    (_, tmp_repo) = mk_repo()
    tmp_session = tmp_repo.writable_session(branch="main")
    tmp_store = tmp_session.store

    root = zarr.group(store=tmp_store)
    array = root.ones(name="ones", shape=(10, 10), chunks=(5, 5), dtype="float32")
    array[:] = 20
    tmp_session.commit("firsttt")

    pickled_store = pickle.dumps(tmp_store)
    store_loaded = pickle.loads(pickled_store)

    root_loaded = zarr.open_group(store_loaded, mode="r")
    array_loaded = root_loaded["ones"]

    assert type(array_loaded) is zarr.Array
    # pickled stores dont point to the same session instance, so they are not equal
    assert array_loaded != array
    assert array_loaded[0, 5] == 20

    pickled_session = pickle.dumps(tmp_session)
    session_loaded = pickle.loads(pickled_session)
    assert tmp_session.snapshot_id == session_loaded.snapshot_id
    assert tmp_session.branch == session_loaded.branch


def test_pickle_error() -> None:
    error = IcechunkError("test")
    pickled = pickle.dumps(error)
    roundtripped = pickle.loads(pickled)
    assert error.message == roundtripped.message
